# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Cleanup task for cleaning up unneeded testcases."""

import collections
import datetime
import json

from googleapiclient.errors import HttpError

from clusterfuzz._internal.base import dates
from clusterfuzz._internal.base import errors
from clusterfuzz._internal.base import memoize
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.chrome import build_info
from clusterfuzz._internal.crash_analysis import crash_comparer
from clusterfuzz._internal.crash_analysis import severity_analyzer
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import ndb_utils
from clusterfuzz._internal.fuzzing import leak_blacklist
from clusterfuzz._internal.metrics import crash_stats
from clusterfuzz._internal.metrics import logs
from handlers import base_handler
from libs import handler
from libs import mail
from libs.issue_management import issue_filer
from libs.issue_management import issue_tracker_policy
from libs.issue_management import issue_tracker_utils

GENERIC_INCORRECT_COMMENT = (
    '\n\nIf this is incorrect, please add the {label_text}')
OSS_FUZZ_INCORRECT_COMMENT = ('\n\nIf this is incorrect, please file a bug on '
                              'https://github.com/google/oss-fuzz/issues/new')

TOP_CRASHES_LIMIT = 5
TOP_CRASHES_DAYS_LOOKBEHIND = 7
TOP_CRASHES_MIN_THRESHOLD = 50 * TOP_CRASHES_DAYS_LOOKBEHIND
TOP_CRASHES_IGNORE_CRASH_TYPES = [
    'Out-of-memory',
    'Stack-overflow',
    'Timeout',
]
TOP_CRASHES_IGNORE_CRASH_STATES = ['NULL']

FUZZ_TARGET_UNUSED_THRESHOLD = 15
UNUSED_HEARTBEAT_THRESHOLD = 15

ProjectMap = collections.namedtuple('ProjectMap', 'jobs platforms')


def _get_predator_result_item(testcase, key, default=None):
  """Return the suspected components for a test case."""
  predator_result = testcase.get_metadata('predator_result')
  if not predator_result:
    return default

  return predator_result['result'].get(key, default)


def _append_generic_incorrect_comment(comment, policy, issue, suffix):
  """Get the generic incorrect comment."""
  wrong_label = policy.label('wrong')
  if not wrong_label:
    return comment

  return comment + GENERIC_INCORRECT_COMMENT.format(
      label_text=issue.issue_tracker.label_text(wrong_label)) + suffix


def job_platform_to_real_platform(job_platform):
  """Get real platform from job platform."""
  for platform in data_types.PLATFORMS:
    if platform in job_platform:
      return platform

  raise ValueError('Unknown platform: ' + job_platform)


def cleanup_reports_metadata():
  """Delete ReportMetadata for uploaded reports."""
  uploaded_reports = ndb_utils.get_all_from_query(
      data_types.ReportMetadata.query(
          ndb_utils.is_true(data_types.ReportMetadata.is_uploaded)),
      keys_only=True)
  ndb_utils.delete_multi(uploaded_reports)


def cleanup_testcases_and_issues():
  """Clean up unneeded open testcases and their associated issues."""
  jobs = data_handler.get_all_job_type_names()
  testcase_keys = ndb_utils.get_all_from_query(
      data_types.Testcase.query(
          ndb_utils.is_false(data_types.Testcase.triaged)),
      keys_only=True)
  top_crashes_by_project_and_platform_map = (
      get_top_crashes_for_all_projects_and_platforms())

  utils.python_gc()

  testcases_processed = 0
  empty_issue_tracker_policy = issue_tracker_policy.get_empty()
  for testcase_key in testcase_keys:
    testcase_id = testcase_key.id()
    try:
      testcase = data_handler.get_testcase_by_id(testcase_id)
    except errors.InvalidTestcaseError:
      # Already deleted.
      continue

    logs.log('Processing testcase %d.' % testcase_id)

    try:
      issue = issue_tracker_utils.get_issue_for_testcase(testcase)
      policy = issue_tracker_utils.get_issue_tracker_policy_for_testcase(
          testcase)
      if not policy:
        policy = empty_issue_tracker_policy

      # Issue updates.
      update_os_labels(policy, testcase, issue)
      update_fuzz_blocker_label(policy, testcase, issue,
                                top_crashes_by_project_and_platform_map)
      update_component_labels(testcase, issue)
      update_issue_ccs_from_owners_file(policy, testcase, issue)
      update_issue_owner_and_ccs_from_predator_results(policy, testcase, issue)
      update_issue_labels_for_flaky_testcase(policy, testcase, issue)

      # Testcase marking rules.
      mark_duplicate_testcase_as_closed_with_no_issue(testcase)
      mark_issue_as_closed_if_testcase_is_fixed(policy, testcase, issue)
      mark_testcase_as_closed_if_issue_is_closed(policy, testcase, issue)
      mark_testcase_as_closed_if_job_is_invalid(testcase, jobs)
      mark_unreproducible_testcase_as_fixed_if_issue_is_closed(testcase, issue)
      mark_unreproducible_testcase_and_issue_as_closed_after_deadline(
          policy, testcase, issue)
      mark_na_testcase_issues_as_wontfix(policy, testcase, issue)

      # Notification, to be done at end after testcase state is updated from
      # previous rules.
      notify_closed_issue_if_testcase_is_open(policy, testcase, issue)
      notify_issue_if_testcase_is_invalid(policy, testcase, issue)
      notify_uploader_when_testcase_is_processed(policy, testcase, issue)

      # Mark testcase as triage complete if both testcase and associated issue
      # are closed. This also need to be done before the deletion rules.
      mark_testcase_as_triaged_if_needed(testcase, issue)

      # Testcase deletion rules.
      delete_unreproducible_testcase_with_no_issue(testcase)
    except Exception:
      logs.log_error('Failed to process testcase %d.' % testcase_id)

    testcases_processed += 1
    if testcases_processed % 100 == 0:
      utils.python_gc()


def cleanup_unused_fuzz_targets_and_jobs():
  """Clean up unused FuzzTarget and FuzzTargetJob entities."""
  last_run_cutoff = utils.utcnow() - datetime.timedelta(
      days=FUZZ_TARGET_UNUSED_THRESHOLD)

  unused_target_jobs = data_types.FuzzTargetJob.query(
      data_types.FuzzTargetJob.last_run < last_run_cutoff)
  valid_target_jobs = data_types.FuzzTargetJob.query(
      data_types.FuzzTargetJob.last_run >= last_run_cutoff)

  to_delete = [t.key for t in unused_target_jobs]

  valid_fuzz_targets = set(t.fuzz_target_name for t in valid_target_jobs)
  for fuzz_target in ndb_utils.get_all_from_model(data_types.FuzzTarget):
    if fuzz_target.fully_qualified_name() not in valid_fuzz_targets:
      to_delete.append(fuzz_target.key)

  ndb_utils.delete_multi(to_delete)


def get_jobs_and_platforms_for_project():
  """Return a map of projects to jobs and platforms map to use for picking top
  crashes."""
  all_jobs = ndb_utils.get_all_from_model(data_types.Job)
  projects_to_jobs_and_platforms = {}
  for job in all_jobs:
    job_environment = job.get_environment()

    # Skip experimental jobs.
    if utils.string_is_true(job_environment.get('EXPERIMENTAL')):
      continue

    # Skip custom binary jobs.
    if (utils.string_is_true(job_environment.get('CUSTOM_BINARY')) or
        job_environment.get('SYSTEM_BINARY_DIR')):
      continue

    # Skip if explicitly excluded using flag.
    if utils.string_is_true(job_environment.get('EXCLUDE_FROM_TOP_CRASHES')):
      continue

    if job.project not in projects_to_jobs_and_platforms:
      projects_to_jobs_and_platforms[job.project] = ProjectMap(set(), set())

    projects_to_jobs_and_platforms[job.project].jobs.add(job.name)
    projects_to_jobs_and_platforms[job.project].platforms.add(
        job_platform_to_real_platform(job.platform))

  return projects_to_jobs_and_platforms


@memoize.wrap(memoize.Memcache(12 * 60 * 60))
def _get_crash_occurrence_platforms_from_crash_parameters(
    crash_type, crash_state, security_flag, project_name, lookbehind_days):
  """Get platforms from crash stats based on crash parameters."""
  last_hour = crash_stats.get_last_successful_hour()
  if not last_hour:
    # No crash stats available, skip.
    return []

  where_clause = ('crash_type = {crash_type} AND '
                  'crash_state = {crash_state} AND '
                  'security_flag = {security_flag} AND '
                  'project = {project}').format(
                      crash_type=json.dumps(crash_type),
                      crash_state=json.dumps(crash_state),
                      security_flag=json.dumps(security_flag),
                      project=json.dumps(project_name),
                  )

  _, rows = crash_stats.get(
      end=last_hour,
      block='day',
      days=lookbehind_days,
      group_by='platform',
      where_clause=where_clause,
      group_having_clause='',
      sort_by='total_count',
      offset=0,
      limit=1)

  platforms = set()
  for row in rows:
    for group in row['groups']:
      platform = group['name'].split(':')[0]
      platforms.add(platform.lower())
  return list(platforms)


def get_platforms_from_testcase_variants(testcase):
  """Get platforms from crash stats based on crash parameters."""
  variant_query = data_types.TestcaseVariant.query(
      data_types.TestcaseVariant.testcase_id == testcase.key.id())
  platforms = {
      variant.platform
      for variant in variant_query
      if variant.is_similar and variant.platform
  }
  return platforms


def get_crash_occurrence_platforms(testcase, lookbehind_days=1):
  """Get platforms from crash stats for a testcase."""
  return set(
      _get_crash_occurrence_platforms_from_crash_parameters(
          testcase.crash_type, testcase.crash_state, testcase.security_flag,
          testcase.project_name, lookbehind_days))


def get_top_crashes_for_all_projects_and_platforms():
  """Return top crashes for all projects and platforms."""
  last_hour = crash_stats.get_last_successful_hour()
  if not last_hour:
    # No crash stats available, skip.
    return {}

  projects_to_jobs_and_platforms = (get_jobs_and_platforms_for_project())
  top_crashes_by_project_and_platform_map = {}

  for project_name, project_map in projects_to_jobs_and_platforms.items():
    top_crashes_by_project_and_platform_map[project_name] = {}

    for platform in project_map.platforms:
      where_clause = (
          'crash_type NOT IN UNNEST(%s) AND '
          'crash_state NOT IN UNNEST(%s) AND '
          'job_type IN UNNEST(%s) AND '
          'platform LIKE %s AND '
          'project = %s' % (json.dumps(TOP_CRASHES_IGNORE_CRASH_TYPES),
                            json.dumps(TOP_CRASHES_IGNORE_CRASH_STATES),
                            json.dumps(list(project_map.jobs)),
                            json.dumps(platform.lower() + '%'),
                            json.dumps(project_name)))

      _, rows = crash_stats.get(
          end=last_hour,
          block='day',
          days=TOP_CRASHES_DAYS_LOOKBEHIND,
          group_by='platform',
          where_clause=where_clause,
          group_having_clause='',
          sort_by='total_count',
          offset=0,
          limit=TOP_CRASHES_LIMIT)
      if not rows:
        continue

      top_crashes_by_project_and_platform_map[project_name][platform] = [{
          'crashState': row['crashState'],
          'crashType': row['crashType'],
          'isSecurity': row['isSecurity'],
          'totalCount': row['totalCount'],
      } for row in rows if row['totalCount'] >= TOP_CRASHES_MIN_THRESHOLD]

  return top_crashes_by_project_and_platform_map


def get_top_crash_platforms(testcase, top_crashes_by_project_and_platform_map):
  """Return list of platforms where this testcase is a top crasher."""
  if testcase.project_name not in top_crashes_by_project_and_platform_map:
    return []

  top_crashes_by_platform_map = top_crashes_by_project_and_platform_map[
      testcase.project_name]
  top_crash_platforms = set()
  for platform in list(top_crashes_by_platform_map.keys()):
    top_crashes = top_crashes_by_platform_map[platform]
    if not top_crashes:
      continue

    for top_crash in top_crashes:
      crash_state_comparer = crash_comparer.CrashComparer(
          top_crash['crashState'], testcase.crash_state)
      crash_type_comparer = crash_comparer.CrashComparer(
          top_crash['crashType'], testcase.crash_type)
      if (crash_state_comparer.is_similar() and
          top_crash['isSecurity'] == testcase.security_flag and
          (top_crash['isSecurity'] or crash_type_comparer.is_similar())):
        top_crash_platforms.add(platform.lower())

  return sorted(list(top_crash_platforms))


def delete_unreproducible_testcase_with_no_issue(testcase):
  """Delete an unreproducible testcase if it has no associated issue and has
  been open for a certain time interval."""
  # Make sure that this testcase is an unreproducible bug. If not, bail out.
  if not testcase.one_time_crasher_flag:
    return

  # Make sure that this testcase has no associated bug. If not, bail out.
  if testcase.bug_information:
    return

  # Make sure that testcase is atleast older than
  # |UNREPRODUCIBLE_TESTCASE_NO_BUG_DEADLINE|, otherwise it will be seen in
  # crash stats anyway.
  if (testcase.timestamp and not dates.time_has_expired(
      testcase.timestamp,
      days=data_types.UNREPRODUCIBLE_TESTCASE_NO_BUG_DEADLINE)):
    return

  # Make sure that testcase is not seen in crash stats for a certain time
  # interval.
  if get_crash_occurrence_platforms(
      testcase, data_types.UNREPRODUCIBLE_TESTCASE_NO_BUG_DEADLINE):
    return

  testcase.key.delete()
  logs.log(
      'Deleted unreproducible testcase %d with no issue.' % testcase.key.id())


def mark_duplicate_testcase_as_closed_with_no_issue(testcase):
  """Closes a duplicate testcase if it has no associated issue and has been open
  for a certain time interval."""
  # Make sure that this testcase is a duplicate bug. If not, bail out.
  if testcase.status != 'Duplicate':
    return

  # Make sure that this testcase has no associated bug. If not, bail out.
  if testcase.bug_information:
    return

  # Make sure that testcase has been open for a certain time interval. We do
  # a null timestamp check since some older testcases could be missing it.
  if (testcase.timestamp and not dates.time_has_expired(
      testcase.timestamp, days=data_types.DUPLICATE_TESTCASE_NO_BUG_DEADLINE)):
    return

  testcase.fixed = 'NA'
  testcase.open = False
  testcase.put()
  logs.log('Closed duplicate testcase %d with no issue.' % testcase.key.id())


def mark_issue_as_closed_if_testcase_is_fixed(policy, testcase, issue):
  """Mark an issue as fixed if all of its associated reproducible testcase are
  fixed."""
  verified_label = policy.label('verified')
  if not verified_label:
    return

  # If there is no associated issue, then bail out.
  if not issue or not testcase.bug_information:
    return

  # If the issue is closed in a status other than Fixed, like Duplicate, WontFix
  # or Archived, we shouldn't change it. Bail out.
  if not issue.is_open and issue.status != policy.status('fixed'):
    return

  # Check testcase status, so as to skip unreproducible uploads.
  if testcase.status not in ['Processed', 'Duplicate']:
    return

  # If the testcase is still open, no work needs to be done. Bail out.
  if testcase.open:
    return

  # FIXME: Find a better solution to skip over reproducible tests that are now
  # showing up a flaky (esp when we are unable to reproduce crash in original
  # crash revision).
  if testcase.fixed == 'NA':
    return

  # We can only verify fixed issues for reproducible testcases. If the testcase
  # is unreproducible, bail out. Exception is if we explicitly marked this as
  # fixed.
  if testcase.one_time_crasher_flag and testcase.fixed != 'Yes':
    return

  # Make sure that no other testcases associated with this issue are open.
  similar_testcase = data_types.Testcase.query(
      data_types.Testcase.bug_information == testcase.bug_information,
      ndb_utils.is_true(data_types.Testcase.open),
      ndb_utils.is_false(data_types.Testcase.one_time_crasher_flag)).get()
  if similar_testcase:
    return

  # As a last check, do the expensive call of actually checking all issue
  # comments to make sure we didn't do the verification already and we didn't
  # get called out on issue mistriage.
  if (issue_tracker_utils.was_label_added(issue, verified_label) or
      issue_tracker_utils.was_label_added(issue, policy.label('wrong'))):
    return

  issue.labels.add(verified_label)
  comment = 'ClusterFuzz testcase %d is verified as fixed' % testcase.key.id()

  fixed_range_url = data_handler.get_fixed_range_url(testcase)
  if fixed_range_url:
    comment += ' in ' + fixed_range_url
  else:
    comment += '.'

  if utils.is_oss_fuzz():
    comment += OSS_FUZZ_INCORRECT_COMMENT
  else:
    comment = _append_generic_incorrect_comment(comment, policy, issue,
                                                ' and re-open the issue.')

  skip_auto_close = data_handler.get_value_from_job_definition(
      testcase.job_type, 'SKIP_AUTO_CLOSE_ISSUE')
  if not skip_auto_close:
    issue.status = policy.status('verified')

  issue.save(new_comment=comment, notify=True)
  logs.log('Mark issue %d as verified for fixed testcase %d.' %
           (issue.id, testcase.key.id()))
  issue_filer.notify_issue_update(testcase, 'verified')


def mark_unreproducible_testcase_as_fixed_if_issue_is_closed(testcase, issue):
  """Mark an unreproducible testcase as fixed if the associated issue is
  closed."""
  # If the testcase is already closed, no more work to do.
  if not testcase.open:
    return

  # Make sure that this testcase is an unreproducible bug. If not, bail out.
  if not testcase.one_time_crasher_flag:
    return

  # Make sure that this testcase has an associated bug. If not, bail out.
  if not testcase.bug_information:
    return

  # Make sure that there is an associated bug and it is in closed state.
  if not issue or issue.is_open:
    return

  testcase.fixed = 'NA'
  testcase.open = False
  testcase.put()
  logs.log('Closed unreproducible testcase %d with issue closed.' %
           testcase.key.id())


def mark_unreproducible_testcase_and_issue_as_closed_after_deadline(
    policy, testcase, issue):
  """Closes an unreproducible testcase and its associated issue after a certain
  time period."""
  # If the testcase is already closed, no more work to do.
  if not testcase.open:
    return

  # Check testcase status, so as to skip unreproducible uploads.
  if testcase.status not in ['Processed', 'Duplicate']:
    return

  # Make sure that this testcase is an unreproducible bug. If not, bail out.
  if not testcase.one_time_crasher_flag:
    return

  # Make sure that this testcase has an associated bug. If not, bail out.
  if not testcase.bug_information:
    return

  # If this testcase was manually uploaded, don't change issue state as our
  # reproduction result might be incorrect.
  if testcase.uploader_email:
    return

  # Make sure that there is an associated bug and it is in open state.
  if not issue or not issue.is_open:
    return

  # Skip closing if flag is set.
  skip_auto_close = data_handler.get_value_from_job_definition(
      testcase.job_type, 'SKIP_AUTO_CLOSE_ISSUE')
  if skip_auto_close:
    return

  # Check if there are any reproducible open testcases are associated with
  # this bug. If yes, return.
  similar_testcase = data_types.Testcase.query(
      data_types.Testcase.bug_information == testcase.bug_information,
      ndb_utils.is_true(data_types.Testcase.open),
      ndb_utils.is_false(data_types.Testcase.one_time_crasher_flag)).get()
  if similar_testcase:
    return

  # Make sure that testcase is atleast older than
  # |UNREPRODUCIBLE_TESTCASE_WITH_BUG_DEADLINE|, otherwise it will be seen in
  # crash stats anyway.
  if (testcase.timestamp and not dates.time_has_expired(
      testcase.timestamp,
      days=data_types.UNREPRODUCIBLE_TESTCASE_WITH_BUG_DEADLINE)):
    return

  # Handle testcase that turned from reproducible to unreproducible. Account
  # for the recent progression task run time.
  last_tested_crash_time = testcase.get_metadata('last_tested_crash_time')
  if (last_tested_crash_time and not dates.time_has_expired(
      last_tested_crash_time,
      days=data_types.UNREPRODUCIBLE_TESTCASE_WITH_BUG_DEADLINE)):
    return

  # Make that there is no crash seen in the deadline period.
  if get_crash_occurrence_platforms(
      testcase, data_types.UNREPRODUCIBLE_TESTCASE_WITH_BUG_DEADLINE):
    return

  # As a last check, do the expensive call of actually checking all issue
  # comments to make sure we we didn't get called out on issue mistriage.
  if issue_tracker_utils.was_label_added(issue, policy.label('wrong')):
    return

  # Close associated issue and testcase.
  comment = ('ClusterFuzz testcase %d is flaky and no longer crashes, '
             'so closing issue.' % testcase.key.id())
  if utils.is_oss_fuzz():
    comment += OSS_FUZZ_INCORRECT_COMMENT
  else:
    comment = _append_generic_incorrect_comment(comment, policy, issue,
                                                ' and re-open the issue.')

  skip_auto_close = data_handler.get_value_from_job_definition(
      testcase.job_type, 'SKIP_AUTO_CLOSE_ISSUE')
  if not skip_auto_close:
    issue.status = policy.status('wontfix')

  issue.save(new_comment=comment, notify=True)
  testcase.fixed = 'NA'
  testcase.open = False
  testcase.put()

  logs.log('Closed unreproducible testcase %d and associated issue.' %
           testcase.key.id())


def mark_na_testcase_issues_as_wontfix(policy, testcase, issue):
  """Mark issues for testcases with fixed == 'NA' as fixed."""
  # Check for for closed, NA testcases.
  if testcase.open or testcase.fixed != 'NA':
    return

  # Nothing to be done if no issue is attached, or if issue is already closed.
  if not issue or not issue.is_open:
    return

  # Make sure that no other testcases associated with this issue are open.
  similar_testcase = data_types.Testcase.query(
      data_types.Testcase.bug_information == testcase.bug_information,
      ndb_utils.is_true(data_types.Testcase.open),
      ndb_utils.is_false(data_types.Testcase.one_time_crasher_flag)).get()
  if similar_testcase:
    return

  # Make that there is no crash seen in the deadline period.
  if get_crash_occurrence_platforms(
      testcase, data_types.UNREPRODUCIBLE_TESTCASE_WITH_BUG_DEADLINE):
    return

  # As a last check, do the expensive call of actually checking all issue
  # comments to make sure we we didn't get called out on issue mistriage.
  if issue_tracker_utils.was_label_added(issue, policy.label('wrong')):
    return

  comment = (f'ClusterFuzz testcase {testcase.key.id()} is closed as invalid, '
             'so closing issue.')

  skip_auto_close = data_handler.get_value_from_job_definition(
      testcase.job_type, 'SKIP_AUTO_CLOSE_ISSUE')
  if not skip_auto_close:
    issue.status = policy.status('wontfix')

  issue.save(new_comment=comment, notify=True)
  logs.log(
      f'Closing issue {issue.id} for invalid testcase {testcase.key.id()}.')


def mark_testcase_as_triaged_if_needed(testcase, issue):
  """Mark testcase as triage complete if both testcase and associated issue
  are closed."""
  # Check if testcase is open. If yes, bail out.
  if testcase.open:
    return

  # Check if there is an associated bug in open state. If yes, bail out.
  if issue:
    # Get latest issue object to ensure our update went through.
    issue = issue_tracker_utils.get_issue_for_testcase(testcase)
    if issue.is_open:
      return

  testcase.triaged = True
  testcase.put()


def mark_testcase_as_closed_if_issue_is_closed(policy, testcase, issue):
  """Mark testcase as closed if the associated issue is closed."""
  # If the testcase is already closed, no more work to do.
  if not testcase.open:
    return

  # If there is no associated issue, then bail out.
  if not issue or not testcase.bug_information:
    return

  # If the issue is still open, no work needs to be done. Bail out.
  if issue.is_open:
    return

  # Make sure we passed our deadline based on issue closed timestamp.
  if (issue.closed_time and not dates.time_has_expired(
      issue.closed_time,
      days=data_types.CLOSE_TESTCASE_WITH_CLOSED_BUG_DEADLINE)):
    return

  # If the issue has an ignore label, don't close the testcase and bail out.
  # This helps to prevent new bugs from getting filed for legit WontFix cases.
  if issue_tracker_utils.was_label_added(issue, policy.label('ignore')):
    return

  testcase.open = False
  testcase.fixed = 'NA'
  testcase.put()
  logs.log('Closed testcase %d with issue closed.' % testcase.key.id())


def mark_testcase_as_closed_if_job_is_invalid(testcase, jobs):
  """Mark testcase as closed if the associated job type does not exist."""
  # If the testcase is already closed, no more work to do.
  if not testcase.open:
    return

  # Check if the testcase job name is in the list of jobs.
  if testcase.job_type in jobs:
    return

  testcase.open = False
  testcase.fixed = 'NA'
  testcase.put()
  logs.log('Closed testcase %d with invalid job.' % testcase.key.id())


def notify_closed_issue_if_testcase_is_open(policy, testcase, issue):
  """Notify closed issue if associated testcase is still open after a certain
  time period."""
  needs_feedback_label = policy.label('needs_feedback')
  if not needs_feedback_label:
    return

  # If the testcase is already closed, no more work to do.
  if not testcase.open:
    return

  # Check testcase status, so as to skip unreproducible uploads.
  if testcase.status not in ['Processed', 'Duplicate']:
    return

  # If there is no associated issue, then bail out.
  if not issue or not testcase.bug_information:
    return

  # If the issue is still open, no work needs to be done. Bail out.
  if issue.is_open:
    return

  # If we have already passed our deadline based on issue closed timestamp,
  # no need to notify. We will close the testcase instead.
  if (issue.closed_time and not dates.time_has_expired(
      issue.closed_time,
      days=data_types.NOTIFY_CLOSED_BUG_WITH_OPEN_TESTCASE_DEADLINE)):
    return

  # Check if there is ignore label on issue already. If yes, bail out.
  if issue_tracker_utils.was_label_added(issue, policy.label('ignore')):
    return

  # Check if we did add the notification comment already. If yes, bail out.
  if issue_tracker_utils.was_label_added(issue, needs_feedback_label):
    return

  issue.labels.add(needs_feedback_label)

  if issue.status in [policy.status('fixed'), policy.status('verified')]:
    issue_comment = (
        'ClusterFuzz testcase {id} is still reproducing on tip-of-tree build '
        '(trunk).\n\nPlease re-test your fix against this testcase and if the '
        'fix was incorrect or incomplete, please re-open the bug.'
    ).format(id=testcase.key.id())

    wrong_label = policy.label('wrong')
    if wrong_label:
      issue_comment += (
          (' Otherwise, ignore this notification and add the '
           '{label_text}.'
          ).format(label_text=issue.issue_tracker.label_text(wrong_label)))
  else:
    # Covers WontFix, Archived cases.
    issue_comment = (
        'ClusterFuzz testcase {id} is still reproducing on tip-of-tree build '
        '(trunk).\n\nIf this testcase was not reproducible locally or '
        'unworkable, ignore this notification and we will file another '
        'bug soon with hopefully a better and workable testcase.\n\n'.format(
            id=testcase.key.id()))
    ignore_label = policy.label('ignore')
    if ignore_label:
      issue_comment += (
          'Otherwise, if this is not intended to be fixed (e.g. this is an '
          'intentional crash), please add the {label_text} to '
          'prevent future bug filing with similar crash stacktrace.'.format(
              label_text=issue.issue_tracker.label_text(ignore_label)))

  issue.save(new_comment=issue_comment, notify=True)
  logs.log('Notified closed issue for open testcase %d.' % testcase.key.id())


def notify_issue_if_testcase_is_invalid(policy, testcase, issue):
  """Leave comments on associated issues when test cases are no longer valid."""
  invalid_fuzzer_label = policy.label('invalid_fuzzer')
  if not invalid_fuzzer_label:
    return

  if not issue or not testcase.bug_information:
    return

  # If the issue is closed, there's no work to do.
  if not issue.is_open:
    return

  # Currently, this only happens if a test case relies on a fuzzer that has
  # been deleted. This can be modified if more cases are needed in the future.
  if not testcase.get_metadata('fuzzer_was_deleted'):
    return

  # Check if we added this message once. If yes, bail out.
  if issue_tracker_utils.was_label_added(issue, invalid_fuzzer_label):
    return

  issue_comment = (
      'ClusterFuzz testcase %d is associated with an obsolete fuzzer and can '
      'no longer be processed. Please close the issue if it is no longer '
      'actionable.') % testcase.key.id()
  issue.labels.add(invalid_fuzzer_label)
  issue.save(new_comment=issue_comment, notify=True)

  logs.log('Closed issue %d for invalid testcase %d.' % (issue.id,
                                                         testcase.key.id()))


def _send_email_to_uploader(testcase_id, to_email, content):
  """Send email to uploader when all the testcase tasks are finished."""
  subject = 'Your testcase upload %d analysis is complete.' % testcase_id
  content_with_footer = (
      '%s\n\n'
      'If you suspect that the result above is incorrect, '
      'try re-doing that job on the testcase report page.') % content.strip()
  html_content = content_with_footer.replace('\n', '<br>')

  mail.send(to_email, subject, html_content)


def _get_severity_from_labels(security_severity_label, labels):
  """Get the severity from the label list."""
  pattern = issue_filer.get_label_pattern(security_severity_label)
  for label in labels:
    match = pattern.match(label)
    if match:
      return severity_analyzer.string_to_severity(match.group(1))

  return data_types.SecuritySeverity.MISSING


def _update_issue_security_severity_and_get_comment(policy, testcase, issue):
  """Apply a new security severity label if none exists on issue already
  and return a comment on this addition. If a label already exists and does
  not match security severity label on issue, then just return a comment on
  what the recommended severity is."""
  security_severity_label = policy.label('security_severity')
  if not security_severity_label:
    return ''

  if not data_types.SecuritySeverity.is_valid(testcase.security_severity):
    return ''

  issue_severity = _get_severity_from_labels(security_severity_label,
                                             issue.labels)

  recommended_severity = issue_filer.apply_substitutions(
      policy, security_severity_label, testcase)
  if not recommended_severity:
    return ''

  recommended_severity = recommended_severity[0]
  if issue_severity == data_types.SecuritySeverity.MISSING:
    issue.labels.add(recommended_severity)
    return ('\n\nA recommended severity was added to this bug. '
            'Please change the severity if it is inaccurate.')
  if issue_severity != testcase.security_severity:
    return (
        '\n\nThe recommended severity (%s) is different from what was assigned '
        'to the bug. Please double check the accuracy of the assigned '
        'severity.' % recommended_severity)

  return ''


def _update_issue_when_uploaded_testcase_is_processed(
    policy, testcase, issue, description, update_bug_summary, notify):
  """Add issue comment when uploaded testcase is processed."""
  if update_bug_summary and testcase.is_crash():
    issue.title = data_handler.get_issue_summary(testcase)

  # Impact labels like impacting head/beta/stable only apply for Chromium.
  if testcase.project_name == 'chromium':
    issue_filer.update_issue_impact_labels(testcase, issue)

  # Add severity labels for all project types.
  comment = description + _update_issue_security_severity_and_get_comment(
      policy, testcase, issue)
  issue.save(new_comment=comment, notify=notify)


def notify_uploader_when_testcase_is_processed(policy, testcase, issue):
  """Notify uploader by email when all the testcase tasks are finished."""
  testcase_id = testcase.key.id()

  # Check if this is a user upload. If not, bail out.
  upload_metadata = data_types.TestcaseUploadMetadata.query(
      data_types.TestcaseUploadMetadata.testcase_id == testcase_id).get()
  if not upload_metadata:
    return

  # Check that we have a valid email to send the notification. If not, bail out.
  to_email = upload_metadata.uploader_email
  if not to_email:
    return

  # If this is a bundled archive with multiple testcases, then don't send email
  # for individual testcases.
  if upload_metadata.bundled:
    return

  # Check if the notification is already sent once. If yes, bail out.
  if data_handler.is_notification_sent(testcase_id, to_email):
    return

  # Make sure all testcase taks are done (e.g. minimization, regression, etc).
  if not data_handler.critical_tasks_completed(testcase):
    return

  notify = not upload_metadata.quiet_flag
  # If the same issue was specified at time of upload, update it.
  if (issue and str(issue.id) == upload_metadata.bug_information and
      not testcase.duplicate_of):
    issue_description = data_handler.get_issue_description(testcase)
    _update_issue_when_uploaded_testcase_is_processed(
        policy, testcase, issue, issue_description,
        upload_metadata.bug_summary_update_flag, notify)

  if notify:
    issue_description_without_crash_state = data_handler.get_issue_description(
        testcase, hide_crash_state=True)
    _send_email_to_uploader(testcase_id, to_email,
                            issue_description_without_crash_state)

  # Make sure to create notification entry, as we use this to update bug.
  data_handler.create_notification_entry(testcase_id, to_email)


def update_os_labels(policy, testcase, issue):
  """Add OS labels to issue."""
  os_label = policy.label('os')
  if not os_label:
    return

  if not issue:
    return

  platforms = get_crash_occurrence_platforms(testcase)
  platforms = platforms.union(get_platforms_from_testcase_variants(testcase))
  logs.log(
      'Found %d platforms for the testcase %d.' % (len(platforms),
                                                   testcase.key.id()),
      platforms=platforms)
  for platform in platforms:
    label = os_label.replace('%PLATFORM%', platform.capitalize())
    if not issue_tracker_utils.was_label_added(issue, label):
      issue.labels.add(label)

  issue.save(notify=False)
  logs.log('Updated labels of issue %d.' % issue.id, labels=issue.labels)


def update_fuzz_blocker_label(policy, testcase, issue,
                              top_crashes_by_project_and_platform_map):
  """Add top crash label to issue."""
  fuzz_blocker_label = policy.label('fuzz_blocker')
  if not fuzz_blocker_label:
    return

  if not issue:
    return

  if not testcase.open:
    return

  top_crash_platforms = get_top_crash_platforms(
      testcase, top_crashes_by_project_and_platform_map)
  if not top_crash_platforms:
    # Not a top crasher, bail out.
    return

  if issue_tracker_utils.was_label_added(issue, fuzz_blocker_label):
    # Issue was already marked a top crasher, bail out.
    return

  if len(top_crash_platforms) == 1:
    platform_message = '%s platform' % top_crash_platforms[0]
  else:
    platform_message = '%s and %s platforms' % (', '.join(
        top_crash_platforms[:-1]), top_crash_platforms[-1])

  fuzzer_name = (
      testcase.get_metadata('fuzzer_binary_name') or testcase.fuzzer_name)
  update_message = (
      'This crash occurs very frequently on %s and is likely preventing the '
      'fuzzer %s from making much progress. Fixing this will allow more bugs '
      'to be found.' % (platform_message, fuzzer_name))
  if utils.is_oss_fuzz():
    update_message += OSS_FUZZ_INCORRECT_COMMENT
  elif utils.is_chromium():
    update_message += '\n\nMarking this bug as a blocker for next Beta release.'
    update_message = _append_generic_incorrect_comment(
        update_message,
        policy,
        issue,
        ' and remove the {label_text}.'.format(
            label_text=issue.issue_tracker.label_text(
                data_types.CHROMIUM_ISSUE_RELEASEBLOCK_BETA_LABEL)))
    issue.labels.add(data_types.CHROMIUM_ISSUE_RELEASEBLOCK_BETA_LABEL)

    # Update with the next beta for trunk, and remove existing milestone label.
    beta_milestone_label = (
        'M-%d' % build_info.get_release_milestone('head', testcase.platform))
    if beta_milestone_label not in issue.labels:
      issue.labels.remove_by_prefix('M-')
      issue.labels.add(beta_milestone_label)

  logs.log(update_message)
  issue.labels.add(fuzz_blocker_label)
  issue.save(new_comment=update_message, notify=True)


def update_component_labels(testcase, issue):
  """Add components to the issue if needed."""
  if not issue:
    return

  components = _get_predator_result_item(
      testcase, 'suspected_components', default=[])

  # Remove components already in issue or whose more specific variants exist.
  filtered_components = []
  for component in components:
    found_component_in_issue = any(
        component == issue_component or issue_component.startswith(component +
                                                                   '>')
        for issue_component in issue.components)
    if not found_component_in_issue:
      filtered_components.append(component)

  if not filtered_components:
    # If there are no new components to add, then we shouldn't make any changes
    # to issue.
    return

  # Don't run on issues we've already applied automatic components to in case
  # labels are removed manually. This may cause issues in the event that we
  # rerun a test case, but it seems like a reasonable tradeoff to avoid spam.
  if issue_tracker_utils.was_label_added(
      issue, data_types.CHROMIUM_ISSUE_PREDATOR_AUTO_COMPONENTS_LABEL):
    return

  for filtered_component in filtered_components:
    issue.components.add(filtered_component)

  issue.labels.add(data_types.CHROMIUM_ISSUE_PREDATOR_AUTO_COMPONENTS_LABEL)
  issue_comment = (
      'Automatically applying components based on crash stacktrace and '
      'information from OWNERS files.\n\n'
      'If this is incorrect, please apply the {label_text}.'.format(
          label_text=issue.issue_tracker.label_text(
              data_types.CHROMIUM_ISSUE_PREDATOR_WRONG_COMPONENTS_LABEL)))
  issue.save(new_comment=issue_comment, notify=True)


def update_issue_ccs_from_owners_file(policy, testcase, issue):
  """Add cc to an issue based on owners list from owners file. This is
  currently applicable to fuzz targets only."""
  auto_cc_label = policy.label('auto_cc_from_owners')
  if not auto_cc_label:
    return

  if not issue or not issue.is_open:
    return

  if testcase.get_metadata('has_issue_ccs_from_owners_file'):
    return

  ccs_list = utils.parse_delimited(
      testcase.get_metadata('issue_owners', ''),
      delimiter=',',
      strip=True,
      remove_empty=True)
  if not ccs_list:
    return

  # If we've assigned the ccs before, it likely means we were incorrect.
  # Don't try again for this particular issue.
  if issue_tracker_utils.was_label_added(issue, auto_cc_label):
    return

  ccs_added = False
  actions = list(issue.actions)
  for cc in ccs_list:
    if cc in issue.ccs:
      continue

    # If cc was previously manually removed from the cc list, we assume that
    # they were incorrectly added. Don't try to add them again.
    cc_was_removed = any(cc in action.ccs.removed for action in actions)
    if cc_was_removed:
      continue

    issue.ccs.add(cc)
    ccs_added = True

  if not ccs_added:
    # Everyone we'd expect to see has already been cced on the issue. No need
    # to spam it with another comment. Also, set the metadata to avoid doing
    # this again.
    testcase.set_metadata('has_issue_ccs_from_owners_file', True)
    return

  issue_comment = (
      'Automatically adding ccs based on OWNERS file / target commit history.')
  if utils.is_oss_fuzz():
    issue_comment += OSS_FUZZ_INCORRECT_COMMENT + '.'
  else:
    issue_comment = _append_generic_incorrect_comment(issue_comment, policy,
                                                      issue, '.')

  issue.labels.add(auto_cc_label)
  issue.save(new_comment=issue_comment, notify=True)


def update_issue_labels_for_flaky_testcase(policy, testcase, issue):
  """Update issue reproducibility label when testcase becomes flaky or
  unreproducible."""
  if not issue or not issue.is_open:
    return

  # If the testcase is reproducible, then no change is needed. Bail out.
  if not testcase.one_time_crasher_flag:
    return

  # Make sure that no other reproducible testcases associated with this issue
  # are open. If yes, no need to update label.
  similar_reproducible_testcase = data_types.Testcase.query(
      data_types.Testcase.bug_information == testcase.bug_information,
      ndb_utils.is_true(data_types.Testcase.open),
      ndb_utils.is_false(data_types.Testcase.one_time_crasher_flag)).get()
  if similar_reproducible_testcase:
    return

  reproducible_label = policy.label('reproducible')
  unreproducible_label = policy.label('unreproducible')
  if not reproducible_label or not unreproducible_label:
    return

  # Make sure that this issue is not already marked Unreproducible.
  if unreproducible_label in issue.labels:
    return

  issue.labels.remove(reproducible_label)
  issue.labels.add(unreproducible_label)
  comment = ('ClusterFuzz testcase {testcase_id} appears to be flaky, '
             'updating reproducibility {label_type}.'.format(
                 testcase_id=testcase.key.id(),
                 label_type=issue.issue_tracker.label_type))
  issue.save(new_comment=comment)


def update_issue_owner_and_ccs_from_predator_results(policy,
                                                     testcase,
                                                     issue,
                                                     only_allow_ccs=False):
  """Assign the issue to an appropriate owner if possible."""
  if not issue or not issue.is_open:
    return

  # If the issue already has an owner, we don't need to update the bug.
  if issue.assignee:
    return

  # If there are more than 3 suspected CLs, we can't be confident in the
  # results. Just skip any sort of notification to CL authors in this case.
  suspected_cls = _get_predator_result_item(testcase, 'suspected_cls')
  if not suspected_cls or len(suspected_cls) > 3:
    return

  # If we've assigned an owner or cc once before, it likely means we were
  # incorrect. Don't try again for this particular issue.
  if (issue_tracker_utils.was_label_added(
      issue, data_types.CHROMIUM_ISSUE_PREDATOR_AUTO_OWNER_LABEL) or
      issue_tracker_utils.was_label_added(
          issue, data_types.CHROMIUM_ISSUE_PREDATOR_AUTO_CC_LABEL)):
    return

  # Validate that the suspected CLs have all of the information we need before
  # continuing. This allows us to assume that they are well-formed later,
  # avoiding any potential exceptions that would interrupt this task.
  for suspected_cl in suspected_cls:
    url = suspected_cl.get('url')
    description = suspected_cl.get('description')
    author = suspected_cl.get('author')
    if not url or not description or not author:
      logs.log_error(
          'Suspected CL for testcase %d is missing required information.' %
          testcase.key.id())
      return

  if len(suspected_cls) == 1 and not only_allow_ccs:
    suspected_cl = suspected_cls[0]

    # If this owner has already been assigned before but has since been removed,
    # don't assign it to them again.
    for action in issue.actions:
      if action.assignee == suspected_cls[0]['author']:
        return

    # We have high confidence for the single-CL case, so we assign the owner.
    issue.labels.add(data_types.CHROMIUM_ISSUE_PREDATOR_AUTO_OWNER_LABEL)
    issue.assignee = suspected_cl['author']
    issue.status = policy.status('assigned')
    issue_comment = (
        'Automatically assigning owner based on suspected regression '
        'changelist %s (%s).\n\n'
        'If this is incorrect, please let us know why and apply the %s '
        'label. If you aren\'t the correct owner for this issue, please '
        'unassign yourself as soon as possible so it can be re-triaged.' %
        (suspected_cl['url'], suspected_cl['description'],
         data_types.CHROMIUM_ISSUE_PREDATOR_WRONG_CL_LABEL))

  else:
    if testcase.get_metadata('has_issue_ccs_from_predator_results'):
      return

    issue_comment = (
        'Automatically adding ccs based on suspected regression changelists:'
        '\n\n')
    ccs_added = False

    for suspected_cl in suspected_cls:
      # Update the comment with the suspected CL, regardless of whether or not
      # we're ccing the author. This might, for example, catch the attention of
      # someone who has already been cced.
      author = suspected_cl['author']
      issue_comment += '%s by %s - %s\n\n' % (suspected_cl['description'],
                                              author, suspected_cl['url'])
      if author in issue.ccs:
        continue

      # If an author has previously been manually removed from the cc list,
      # we assume they were incorrectly added. Don't try to add them again.
      author_was_removed = False
      for action in issue.actions:
        if author in action.ccs.removed:
          author_was_removed = True
          break

      if author_was_removed:
        continue

      issue.ccs.add(author)
      ccs_added = True

    if not ccs_added:
      # Everyone we'd expect to see has already been cced on the issue. No need
      # to spam it with another comment. Also, set the metadata to avoid doing
      # this again.
      testcase.set_metadata('has_issue_ccs_from_owners_file', True)
      return

    issue.labels.add(data_types.CHROMIUM_ISSUE_PREDATOR_AUTO_CC_LABEL)
    issue_comment += ((
        'If this is incorrect, please let us know why and apply the '
        '{label_text}.').format(
            label_text=issue.issue_tracker.label_text(
                data_types.CHROMIUM_ISSUE_PREDATOR_WRONG_CL_LABEL)))

  try:
    issue.save(new_comment=issue_comment, notify=True)
  except HttpError:
    # If we see such an error when we aren't setting an owner, it's unexpected.
    if only_allow_ccs or not issue.assignee:
      logs.log_error(
          'Unable to update issue for test case %d.' % testcase.key.id())
      return

    # Retry without setting the owner. They may not be a chromium project
    # member, in which case we can try falling back to cc.
    issue = issue_tracker_utils.get_issue_for_testcase(testcase)
    update_issue_owner_and_ccs_from_predator_results(
        policy, testcase, issue, only_allow_ccs=True)


def cleanup_unused_heartbeats():
  """Clean up unused heartbeat entities."""
  cutoff_time = utils.utcnow() - datetime.timedelta(
      days=UNUSED_HEARTBEAT_THRESHOLD)
  unused_heartbeats = ndb_utils.get_all_from_query(
      data_types.Heartbeat.query(
          data_types.Heartbeat.last_beat_time < cutoff_time),
      keys_only=True)

  ndb_utils.delete_multi(unused_heartbeats)


class Handler(base_handler.Handler):
  """Cleanup."""

  @handler.cron()
  def get(self):
    cleanup_testcases_and_issues()
    cleanup_reports_metadata()
    leak_blacklist.cleanup_global_blacklist()
    cleanup_unused_fuzz_targets_and_jobs()
    cleanup_unused_heartbeats()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the automatic weight adjustment cron job."""
# pylint: disable=protected-access
import datetime
import unittest

import six

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers.cron import fuzzer_and_job_weights

_TEST_SPECIFICATION = fuzzer_and_job_weights.QuerySpecification(
    query_format='ignored',
    formatter=fuzzer_and_job_weights._past_day_formatter,
    reason='matches test specification')


class TestFormatters(unittest.TestCase):
  """Tests for the query formatter functions used by
  fuzzer_and_job_weights.py."""

  def setUp(self):
    test_helpers.patch(self, [
        'clusterfuzz._internal.base.utils.utcnow',
    ])
    self.mock.utcnow.return_value = datetime.datetime(2018, 1, 1)

  def test_past_day_formatter(self):
    """Tests for _past_day_formatter."""
    expected_query = """
SELECT
  fuzzer,
  job,
  1.0 - (1.0 - 0.25) * AVG(field) AS new_weight
FROM
  engine.TestcaseRun
WHERE
  _PARTITIONTIME BETWEEN TIMESTAMP('2017-12-31')
  AND TIMESTAMP('2018-01-01')
GROUP BY
  fuzzer,
  job
"""
    specific_query_format = fuzzer_and_job_weights.GENERIC_QUERY_FORMAT.format(
        field_name='field', min_weight=0.25)
    actual_query = fuzzer_and_job_weights._past_day_formatter(
        specific_query_format, 'engine')
    self.assertEqual(actual_query, expected_query)

  def test_new_fuzzer_formatter(self):
    """Tests for _new_fuzzer_formatter."""
    expected_query = """
SELECT
  fuzzer,
  job,
  5.0 as new_weight,
  MIN(_PARTITIONTIME) as first_time
FROM
  engine.TestcaseRun
GROUP BY
  fuzzer,
  job
HAVING
  first_time >= TIMESTAMP('2017-12-25')
"""
    actual_query = fuzzer_and_job_weights._new_fuzzer_formatter(
        fuzzer_and_job_weights.NEW_FUZZER_FORMAT, 'engine')
    self.assertEqual(actual_query, expected_query)

  def test_coverage_formatter(self):
    """Tests for _coverage_formatter."""
    expected_query = """
SELECT
  recent.fuzzer AS fuzzer,
  recent.job AS job,
  0.75 as new_weight
FROM (
  SELECT
    fuzzer,
    job,
    MAX(edge_coverage / edges_total) AS coverage
  FROM
    engine.TestcaseRun
  WHERE
    _PARTITIONTIME BETWEEN TIMESTAMP('2017-12-24')
    AND TIMESTAMP('2017-12-31')
    AND edges_total > 0
    AND edge_coverage > 0
  GROUP BY
    fuzzer,
    job
  HAVING
    coverage <= 1.0) AS recent
JOIN (
  SELECT
    fuzzer,
    job,
    MAX(edge_coverage / edges_total) AS coverage
  FROM
    engine.TestcaseRun
  WHERE
    _PARTITIONTIME BETWEEN TIMESTAMP('2017-12-17')
    AND TIMESTAMP('2017-12-24')
    AND edges_total > 0
    AND edge_coverage > 0
  GROUP BY
    fuzzer,
    job
  HAVING
    coverage <= 1.0) AS older
ON
  recent.fuzzer = older.fuzzer
  AND recent.job = older.job
WHERE
  ABS((recent.coverage - older.coverage) / recent.coverage) < 0.01
"""
    actual_query = fuzzer_and_job_weights._coverage_formatter(
        fuzzer_and_job_weights.COVERAGE_UNCHANGED_FORMAT, 'engine')
    self.assertEqual(actual_query, expected_query)


@test_utils.with_cloud_emulators('datastore')
class TestUpdateChildWeightsForParentFuzzer(unittest.TestCase):
  """Tests for update_target_weights_for_engine."""

  def setUp(self):
    test_helpers.patch_environ(self)
    test_helpers.patch(self, [
        'handlers.cron.fuzzer_and_job_weights._query_helper',
        'handlers.cron.fuzzer_and_job_weights.'
        'store_current_weights_in_bigquery',
        'handlers.cron.fuzzer_and_job_weights.update_weight_for_target',
    ])

  def test_reported_fuzzer_has_weight_restored(self):
    """Ensure that a target reported fixed has its weight restored."""
    data_types.FuzzTarget(
        engine='libFuzzer', binary='good_fuzzer', project='test-project').put()
    data_types.FuzzTargetJob(
        fuzz_target_name='libFuzzer_good_fuzzer',
        engine='libFuzzer',
        job='asan',
        weight=0.1).put()

    # Report that the issue is corrected.
    self.mock._query_helper.return_value = [
        {
            'fuzzer': 'libFuzzer_good_fuzzer',
            'job': 'asan',
            'new_weight': 1.00,
        },
    ]

    fuzzer_and_job_weights.update_target_weights_for_engine(
        None, 'libFuzzer', [_TEST_SPECIFICATION])
    self.mock.update_weight_for_target.assert_called_with(
        'libFuzzer_good_fuzzer', 'asan',
        fuzzer_and_job_weights.RESTORE_DEFAULT_MATCH)

  def test_weight_increase(self):
    """Ensure that weight increases are possible."""
    data_types.FuzzTarget(
        engine='libFuzzer', binary='very_good_fuzzer',
        project='test-project').put()
    data_types.FuzzTargetJob(
        fuzz_target_name='libFuzzer_very_good_fuzzer',
        engine='libFuzzer',
        job='asan',
        weight=1.0).put()

    # Report that the issue is corrected.
    self.mock._query_helper.return_value = [
        {
            'fuzzer': 'libFuzzer_very_good_fuzzer',
            'job': 'asan',
            'new_weight': 2.00,
        },
    ]

    specification = fuzzer_and_job_weights.QuerySpecification(
        query_format='ignored',
        formatter=fuzzer_and_job_weights._past_day_formatter,
        reason='increase weight for test')
    match = fuzzer_and_job_weights.SpecificationMatch(
        new_weight=2.0, reason=specification.reason)
    fuzzer_and_job_weights.update_target_weights_for_engine(
        None, 'libFuzzer', [specification])
    self.mock.update_weight_for_target.assert_called_with(
        'libFuzzer_very_good_fuzzer', 'asan', match)

  def test_target_ignored_if_not_ran(self):
    """Ensure that we don't reset a target weight if it did not run."""
    data_types.FuzzTarget(
        engine='libFuzzer', binary='good_fuzzer', project='test-project').put()
    data_types.FuzzTargetJob(
        fuzz_target_name='libFuzzer_good_fuzzer',
        engine='libFuzzer',
        job='asan',
        weight=0.1).put()

    # Do not report any runs.
    self.mock._query_helper.return_value = []

    fuzzer_and_job_weights.update_target_weights_for_engine(
        None, 'libFuzzer', [_TEST_SPECIFICATION])
    self.assertFalse(self.mock.update_weight_for_target.called)

  def test_problem_penalized(self):
    """Ensure that we penalize a target for having problems."""
    data_types.FuzzTarget(
        engine='libFuzzer', binary='problematic_fuzzer',
        project='test-project').put()
    data_types.FuzzTargetJob(
        fuzz_target_name='libFuzzer_problematic_fuzzer',
        engine='libFuzzer',
        job='dummy_job',
        weight=0.1).put()

    self.mock._query_helper.return_value = [
        {
            'fuzzer': 'libFuzzer_problematic_fuzzer',
            'job': 'dummy_job',
            'new_weight': 0.25,
        },
    ]

    fuzzer_and_job_weights.update_target_weights_for_engine(
        None, 'libFuzzer', [_TEST_SPECIFICATION])
    expected_match = fuzzer_and_job_weights.SpecificationMatch(
        new_weight=0.25, reason=_TEST_SPECIFICATION.reason)
    self.mock.update_weight_for_target.assert_called_with(
        'libFuzzer_problematic_fuzzer', 'dummy_job', expected_match)

  def test_new_fuzzer(self):
    """Tests to ensure that the new fuzzer query works properly."""
    data_types.FuzzTarget(
        engine='libFuzzer', binary='old_fuzzer', project='test-project').put()
    data_types.FuzzTargetJob(
        fuzz_target_name='libFuzzer_old_fuzzer',
        engine='libFuzzer',
        job='dummy_job',
        weight=1.0).put()

    data_types.FuzzTarget(
        engine='libFuzzer', binary='new_fuzzer', project='test-project').put()
    data_types.FuzzTargetJob(
        fuzz_target_name='libFuzzer_new_fuzzer',
        engine='libFuzzer',
        job='dummy_job',
        weight=1.0).put()

    self.mock._query_helper.return_value = [
        {
            'fuzzer': 'libFuzzer_new_fuzzer',
            'job': 'dummy_job',
            'new_weight': 5.0,
            'first_time': '<irrelavent for test>',
        },
    ]

    fuzzer_and_job_weights.update_target_weights_for_engine(
        None, 'libFuzzer', [fuzzer_and_job_weights.NEW_FUZZER_SPECIFICATION])
    expected_match = fuzzer_and_job_weights.SpecificationMatch(
        new_weight=5.0,
        reason=fuzzer_and_job_weights.NEW_FUZZER_SPECIFICATION.reason)
    self.mock.update_weight_for_target.assert_called_with(
        'libFuzzer_new_fuzzer', 'dummy_job', expected_match)


@test_utils.with_cloud_emulators('datastore')
class TestUpdateJobWeights(unittest.TestCase):
  """Test updating job weights."""

  def setUp(self):
    test_fuzzer_jobs = {
        'libFuzzer': [
            'libfuzzer_asan_job',
            'libfuzzer_asan_large_job',
            'libfuzzer_msan_job',
            'libfuzzer_ubsan_job',
            'libfuzzer_tsan_job',
            'libfuzzer_cfi_job',
            'libfuzzer_asan_job2',
        ],
        'afl': ['afl_asan_job',],
        'honggfuzz': ['honggfuzz_asan_job'],
        'blackbox': ['asan_blackbox_job',]
    }

    for fuzzer, jobs in six.iteritems(test_fuzzer_jobs):
      for job in jobs:
        data_types.Job(name=job).put()
        data_types.FuzzerJob(fuzzer=fuzzer, job=job).put()

    data_types.FuzzTargetsCount(id='libfuzzer_asan_job', count=10).put()
    data_types.FuzzTargetsCount(id='libfuzzer_msan_job', count=5).put()
    data_types.FuzzTargetsCount(id='libfuzzer_ubsan_job', count=5).put()
    data_types.FuzzTargetsCount(id='libfuzzer_tsan_job', count=5).put()
    data_types.FuzzTargetsCount(id='libfuzzer_cfi_job', count=5).put()
    data_types.FuzzTargetsCount(id='afl_asan_job', count=10).put()
    data_types.FuzzTargetsCount(id='libfuzzer_asan_job2', count=0).put()
    data_types.FuzzTargetsCount(id='honggfuzz_asan_job', count=10).put()
    data_types.FuzzTargetsCount(id='libfuzzer_asan_large_job', count=1000).put()

  def test_update_job_weights(self):
    """Test update job weights."""
    fuzzer_and_job_weights.update_job_weights()

    def get_result(job):
      return data_types.FuzzerJob.query(data_types.FuzzerJob.job == job).get()

    self.assertEqual(5.0, get_result('libfuzzer_asan_job').multiplier)
    self.assertEqual(1.0, get_result('libfuzzer_msan_job').multiplier)
    self.assertEqual(0.5, get_result('libfuzzer_ubsan_job').multiplier)
    self.assertEqual(0.5, get_result('libfuzzer_tsan_job').multiplier)
    self.assertEqual(0.5, get_result('libfuzzer_cfi_job').multiplier)
    self.assertEqual(5.0, get_result('afl_asan_job').multiplier)
    self.assertEqual(15.0, get_result('asan_blackbox_job').multiplier)
    self.assertEqual(15.0, get_result('libfuzzer_asan_job2').multiplier)
    self.assertEqual(1.0, get_result('honggfuzz_asan_job').multiplier)
    self.assertEqual(50.0, get_result('libfuzzer_asan_large_job').multiplier)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""commands tests."""
import datetime
import os
import unittest

from google.cloud import ndb
import mock

from clusterfuzz._internal.base import errors
from clusterfuzz._internal.bot.tasks import commands
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils


@commands.set_task_payload
def dummy(_):
  """A dummy function."""
  return os.environ['TASK_PAYLOAD']


@commands.set_task_payload
def dummy_exception(_):
  """A dummy function."""
  raise Exception(os.environ['TASK_PAYLOAD'])


class SetTaskPayloadTest(unittest.TestCase):
  """Test set_task_payload."""

  def setUp(self):
    helpers.patch_environ(self)

  def test_set(self):
    """Test set."""
    task = mock.Mock()
    task.payload.return_value = 'payload something'
    self.assertEqual('payload something', dummy(task))
    self.assertIsNone(os.getenv('TASK_PAYLOAD'))

  def test_exc(self):
    """Test when exception occurs."""
    task = mock.Mock()
    task.payload.return_value = 'payload something'
    with self.assertRaises(Exception) as cm:
      self.assertEqual('payload something', dummy_exception(task))
    self.assertEqual('payload something', str(cm.exception))
    self.assertEqual({'task_payload': 'payload something'}, cm.exception.extras)
    self.assertIsNone(os.getenv('TASK_PAYLOAD'))


@test_utils.with_cloud_emulators('datastore')
class RunCommandTest(unittest.TestCase):
  """Tests for run_command."""

  def setUp(self):
    helpers.patch_environ(self)
    helpers.patch(self, [
        ('fuzz_execute_task',
         'clusterfuzz._internal.bot.tasks.fuzz_task.execute_task'),
        ('progression_execute_task',
         'clusterfuzz._internal.bot.tasks.progression_task.execute_task'),
        'clusterfuzz._internal.base.utils.utcnow',
    ])

    os.environ['BOT_NAME'] = 'bot_name'
    os.environ['TASK_LEASE_SECONDS'] = '60'
    os.environ['FAIL_WAIT'] = '60'
    self.mock.utcnow.return_value = test_utils.CURRENT_TIME

  def test_run_command_fuzz(self):
    """Test run_command with a normal command."""
    commands.run_command('fuzz', 'fuzzer', 'job')

    self.assertEqual(1, self.mock.fuzz_execute_task.call_count)
    self.mock.fuzz_execute_task.assert_called_with('fuzzer', 'job')

    # Fuzz task should not create any TaskStatus entities.
    task_status_entities = list(data_types.TaskStatus.query())
    self.assertEqual(0, len(task_status_entities))

  def test_run_command_progression(self):
    """Test run_command with a progression task."""
    commands.run_command('progression', '123', 'job')

    self.assertEqual(1, self.mock.progression_execute_task.call_count)
    self.mock.progression_execute_task.assert_called_with('123', 'job')

    # TaskStatus should indicate success.
    task_status_entities = list(data_types.TaskStatus.query())
    self.assertEqual(1, len(task_status_entities))

    task_status = task_status_entities[0]
    self.assertEqual(
        ndb.Key(data_types.TaskStatus, 'progression 123 job'), task_status.key)

    self.assertDictEqual({
        'bot_name': 'bot_name',
        'status': 'finished',
        'time': test_utils.CURRENT_TIME,
    }, task_status.to_dict())

  def test_run_command_exception(self):
    """Test run_command with an exception."""
    self.mock.progression_execute_task.side_effect = Exception

    with self.assertRaises(Exception):
      commands.run_command('progression', '123', 'job')

    # TaskStatus should indicate failure.
    task_status_entities = list(data_types.TaskStatus.query())
    self.assertEqual(1, len(task_status_entities))

    task_status = task_status_entities[0]
    self.assertDictEqual({
        'bot_name': 'bot_name',
        'status': 'errored out',
        'time': test_utils.CURRENT_TIME,
    }, task_status.to_dict())

  def test_run_command_invalid_testcase(self):
    """Test run_command with an invalid testcase exception."""
    self.mock.progression_execute_task.side_effect = errors.InvalidTestcaseError
    commands.run_command('progression', '123', 'job')

    task_status_entities = list(data_types.TaskStatus.query())
    self.assertEqual(1, len(task_status_entities))

    # TaskStatus should still indicate success.
    task_status = task_status_entities[0]
    self.assertDictEqual({
        'bot_name': 'bot_name',
        'status': 'finished',
        'time': test_utils.CURRENT_TIME,
    }, task_status.to_dict())

  def test_run_command_already_running(self):
    """Test run_command with another instance currently running."""
    data_types.TaskStatus(
        id='progression 123 job',
        bot_name='another_bot',
        time=test_utils.CURRENT_TIME,
        status='started').put()

    with self.assertRaises(commands.AlreadyRunningError):
      commands.run_command('progression', '123', 'job')

    self.assertEqual(0, self.mock.progression_execute_task.call_count)

    task_status_entities = list(data_types.TaskStatus.query())
    self.assertEqual(1, len(task_status_entities))

    task_status = task_status_entities[0]
    self.assertDictEqual({
        'bot_name': 'another_bot',
        'status': 'started',
        'time': test_utils.CURRENT_TIME,
    }, task_status.to_dict())

  def test_run_command_already_running_expired(self):
    """Test run_command with another instance currently running, but its lease
    has expired."""
    data_types.TaskStatus(
        id='progression 123 job',
        bot_name='another_bot',
        time=datetime.datetime(1970, 1, 1),
        status='started').put()

    commands.run_command('progression', '123', 'job')
    self.assertEqual(1, self.mock.progression_execute_task.call_count)

    task_status_entities = list(data_types.TaskStatus.query())
    self.assertEqual(1, len(task_status_entities))

    task_status = task_status_entities[0]
    self.assertDictEqual({
        'bot_name': 'bot_name',
        'status': 'finished',
        'time': test_utils.CURRENT_TIME,
    }, task_status.to_dict())


class UpdateEnvironmentForJobTest(unittest.TestCase):
  """update_environment_for_job tests."""

  def setUp(self):
    helpers.patch_environ(self)

  def test_basic(self):
    """Basic tests."""
    commands.update_environment_for_job('FUZZ_TEST_TIMEOUT = 123\n'
                                        'MAX_TESTCASES = 5\n'
                                        'B = abcdef\n')
    self.assertEqual(123, environment.get_value('FUZZ_TEST_TIMEOUT'))
    self.assertEqual(5, environment.get_value('MAX_TESTCASES'))
    self.assertEqual('abcdef', environment.get_value('B'))

  def test_timeout_overrides(self):
    """Test timeout overrides."""
    environment.set_value('FUZZ_TEST_TIMEOUT_OVERRIDE', 9001)
    environment.set_value('MAX_TESTCASES_OVERRIDE', 42)
    commands.update_environment_for_job(
        'FUZZ_TEST_TIMEOUT = 123\nMAX_TESTCASES = 5\n')
    self.assertEqual(9001, environment.get_value('FUZZ_TEST_TIMEOUT'))
    self.assertEqual(42, environment.get_value('MAX_TESTCASES'))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Credentials management."""

from oauth2client.client import Credentials
from oauth2client.client import Storage

from clusterfuzz._internal.config import db_config


class CredentialStorage(Storage):
  """Instead of reading a file, just parse a config entry."""

  def locked_get(self):
    """Return Credentials."""
    content = db_config.get_value('client_credentials')
    if not content:
      return None

    try:
      credentials = Credentials.new_from_json(content)
      credentials.set_store(self)
    except ValueError:
      return None

    return credentials

  def locked_put(self, credentials):  # pylint: disable=unused-argument
    pass

  def locked_delete(self):
    pass



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper functions to update fuzzer-job mappings, and select fuzzers to run."""

import collections

from google.cloud import ndb

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import fuzz_target_utils
from clusterfuzz._internal.datastore import ndb_utils
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment

# Used to prepare targets to be passed to utils.random_weighted_choice.
WeightedTarget = collections.namedtuple('WeightedTarget', ['target', 'weight'])


def update_mappings_for_fuzzer(fuzzer, mappings=None):
  """Clear existing mappings for a fuzzer, and replace them."""
  if mappings is None:
    mappings = fuzzer.jobs

  query = data_types.FuzzerJob.query()
  query = query.filter(data_types.FuzzerJob.fuzzer == fuzzer.name)
  entities = ndb_utils.get_all_from_query(query)
  old_mappings = {}
  for entity in entities:
    old_mappings[entity.job] = entity

  new_mappings = []
  for job_name in mappings:
    mapping = old_mappings.pop(job_name, None)
    if mapping:
      continue

    job = data_types.Job.query(data_types.Job.name == job_name).get()
    if not job:
      logs.log_error('An unknown job %s was selected for fuzzer %s.' %
                     (job_name, fuzzer.name))
      continue

    mapping = data_types.FuzzerJob()
    mapping.fuzzer = fuzzer.name
    mapping.job = job_name
    mapping.platform = job.platform
    new_mappings.append(mapping)

  ndb_utils.put_multi(new_mappings)
  ndb_utils.delete_multi([m.key for m in list(old_mappings.values())])


def update_mappings_for_job(job, mappings):
  """Clear existing mappings for a job, and replace them."""
  existing_fuzzers_query = data_types.Fuzzer.query(
      data_types.Fuzzer.jobs == job.name)
  existing_fuzzers = {fuzzer.name: fuzzer for fuzzer in existing_fuzzers_query}
  modified_fuzzers = []

  for fuzzer_name in mappings:
    fuzzer = existing_fuzzers.pop(fuzzer_name, None)
    if fuzzer:
      continue

    fuzzer = data_types.Fuzzer.query(
        data_types.Fuzzer.name == fuzzer_name).get()
    if not fuzzer:
      logs.log_error('An unknown fuzzer %s was selected for job %s.' %
                     (fuzzer_name, job.name))
      continue

    fuzzer.jobs.append(job.name)
    modified_fuzzers.append(fuzzer)
    update_mappings_for_fuzzer(fuzzer)

  # Removing the remaining values in exisiting_fuzzers as
  # they are no longer mapped.
  for fuzzer in existing_fuzzers.values():
    fuzzer.jobs.remove(job.name)
    modified_fuzzers.append(fuzzer)
    update_mappings_for_fuzzer(fuzzer)
  ndb.put_multi(modified_fuzzers)


def update_platform_for_job(job_name, new_platform):
  """Update platform for all mappings for a particular job."""
  query = data_types.FuzzerJob.query()
  query = query.filter(data_types.FuzzerJob.job == job_name)
  mappings = ndb_utils.get_all_from_query(query)
  new_mappings = []
  for mapping in mappings:
    mapping.platform = new_platform
    new_mappings.append(mapping)
  ndb_utils.put_multi(new_mappings)


def get_fuzz_task_payload(platform=None):
  """Select a fuzzer that can run on this platform."""
  if not platform:
    queue_override = environment.get_value('QUEUE_OVERRIDE')
    platform = queue_override if queue_override else environment.platform()

  if environment.is_local_development():
    query = data_types.FuzzerJob.query()
    query = query.filter(data_types.FuzzerJobs.platform == platform)
    mappings = list(ndb_utils.get_all_from_query(query))
  else:
    query = data_types.FuzzerJobs.query()
    query = query.filter(data_types.FuzzerJobs.platform == platform)

    mappings = []
    for entity in query:
      mappings.extend(entity.fuzzer_jobs)

  if not mappings:
    return None, None

  selection = utils.random_weighted_choice(
      mappings, weight_attribute='actual_weight')
  return selection.fuzzer, selection.job


def select_fuzz_target(targets, target_weights):
  """Select a fuzz target from a list of potential targets."""
  assert targets

  weighted_targets = []
  for target in targets:
    weight = target_weights.get(target, 1.0)
    weighted_targets.append(WeightedTarget(target, weight))

  return utils.random_weighted_choice(weighted_targets).target


def get_fuzz_target_weights():
  """Get a list of fuzz target weights based on the current fuzzer."""
  job_type = environment.get_value('JOB_NAME')

  target_jobs = list(fuzz_target_utils.get_fuzz_target_jobs(job=job_type))
  fuzz_targets = fuzz_target_utils.get_fuzz_targets_for_target_jobs(target_jobs)

  weights = {}
  for fuzz_target, target_job in zip(fuzz_targets, target_jobs):
    if not fuzz_target:
      logs.log_error('Skipping weight assignment for fuzz target %s.' %
                     target_job.fuzz_target_name)
      continue

    weights[fuzz_target.binary] = target_job.weight

  return weights



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Mock config."""


class MockConfig(object):
  """Mock config."""

  def __init__(self, data):
    self._data = data

  def get(self, key_name='', default=None):
    """Get key value using a key name."""
    parts = key_name.split('.')
    value = self._data
    for part in parts:
      if part not in value:
        return default

      value = value[part]

    return value

  def sub_config(self, path):
    return MockConfig(self.get(path))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A cron handler that builds data_types.CrashStatistic.

Given `end_hour`, we build the aggregated row for the end hour.

The rows are aggregated by crash_signature, fuzzer_name, job_type, platform,
revision, and project.

Here's how it works:
1. Get the latest end_hour, so that we can run the next one.
2. Send query to BigQuery and tell BigQuery to insert the result to a specific
  table.
3. Poll the result.
4. Store the end_hour.
"""

import datetime
import json
import logging
import time

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.google_cloud_utils import big_query
from clusterfuzz._internal.metrics import crash_stats
from handlers import base_handler
from libs import handler

# After insertion, it takes a few seconds for a record to show up.
# We give it a few minutes.
#
# We add one hour because our hour spans from 0 minutes to 59 minutes.
BIGQUERY_INSERTION_DELAY = datetime.timedelta(hours=1, minutes=2)

# The template for job ids. Running on a already-run hour is fine to certain
# degree. We de-dup when reading.
JOB_ID_TEMPLATE = 'build_crash_stats_test_{unique_number}'

TIMEOUT = 2 * 60

SQL = """
SELECT
    COUNT(*) as count,
    crash_type, crash_state, security_flag, parent_fuzzer_name, fuzzer_name,
    job_type, revision, platform, project, reproducible_flag,
    IF(STARTS_WITH(platform, 'android'), 'android', '') AS parent_platform,
    CAST(FLOOR(UNIX_SECONDS(created_at) / 3600) AS INT64) as hour,
    MIN(crash_time_in_ms) AS min_crash_time_in_ms,
    MAX(crash_time_in_ms) AS max_crash_time_in_ms,
    SUM(crash_time_in_ms) AS sum_crash_time_in_ms,
    CAST(SUM(POW(crash_time_in_ms, 2)) AS INT64) AS sum_square_crash_time_in_ms,
    ANY_VALUE(new_flag=True) AS new_flag
FROM main.crashes
WHERE
  CAST(FLOOR(UNIX_SECONDS(created_at) / 3600) AS INT64) = {end_hour} AND
  _PARTITIONTIME = TIMESTAMP_TRUNC('{end_date}', DAY)
GROUP BY
  crash_type, crash_state, security_flag, parent_fuzzer_name, fuzzer_name,
  job_type, revision, parent_platform, platform, project, hour,
  reproducible_flag
"""


class TooEarlyException(Exception):
  """The end hour is too early according to BIGQUERY_INSERTION_DELAY."""


def get_start_hour():
  """Get the start hour from the first crash."""
  client = big_query.Client()

  sql = """
SELECT min(CAST(FLOOR(UNIX_SECONDS(created_at) / 3600) AS INT64)) as min_hour
FROM main.crashes
"""

  result = client.query(query=sql)
  if result and result.rows:
    return result.rows[0]['min_hour']

  return 0


def get_last_successful_hour_or_start_hour():
  """Get the last hour that ran successfully or the start hour."""
  last_hour = crash_stats.get_last_successful_hour()
  if last_hour:
    return last_hour

  return get_start_hour()


def get_next_end_hour():
  """Get the next end hour. If it's too early to compute data for the next end
    hour, return None."""
  last_successful_hour = get_last_successful_hour_or_start_hour()
  if not last_successful_hour:
    # No crashes seen, too early to start building stats.
    raise TooEarlyException()

  next_end_hour = last_successful_hour + 1

  next_datetime = crash_stats.get_datetime(next_end_hour)
  if (utils.utcnow() - next_datetime) <= BIGQUERY_INSERTION_DELAY:
    raise TooEarlyException()

  return next_end_hour


def make_request(client, job_id, end_hour):
  """Make a request to BigQuery to build crash stats."""
  table_id = (
      'crash_stats$%s' % crash_stats.get_datetime(end_hour).strftime('%Y%m%d'))

  sql = SQL.format(
      end_hour=end_hour,
      end_date=(crash_stats.get_datetime(end_hour).strftime('%Y-%m-%d')))
  logging.info('TableID: %s\nJobID: %s\nSQL: %s', table_id, job_id, sql)

  client.insert_from_query(
      dataset_id='main', table_id=table_id, job_id=job_id, query=sql)


def build(end_hour):
  """Build crash stats for the end hour."""
  logging.info('Started building crash stats for %s.',
               crash_stats.get_datetime(end_hour))
  job_id = JOB_ID_TEMPLATE.format(unique_number=int(time.time()))

  client = big_query.Client()
  make_request(client, job_id, end_hour)

  start_time = time.time()
  while (time.time() - start_time) < TIMEOUT:
    time.sleep(10)

    result = client.get_job(job_id)
    logging.info('Checking %s', json.dumps(result))

    if result['status']['state'] == 'DONE':
      if result['status'].get('errors'):
        raise Exception(json.dumps(result))
      return

  raise Exception('Building crash stats exceeded %d seconds.' % TIMEOUT)


def build_if_needed():
  """Get the next end hour and decide whether to execute build(). If build()
    succeeds, then record the next end hour."""
  try:
    end_hour = get_next_end_hour()
    build(end_hour)

    job_history = data_types.BuildCrashStatsJobHistory()
    job_history.end_time_in_hours = end_hour
    job_history.put()
    logging.info('CrashStatistics for end_hour=%s is built successfully',
                 crash_stats.get_datetime(end_hour))
    return end_hour
  except TooEarlyException:
    logging.info("Skip building crash stats because it's too early.")

  return None


class Handler(base_handler.Handler):
  """Handler for building data_types.CrashsStats2."""

  @handler.cron()
  def get(self):
    """Process a GET request from a cronjob."""
    end_hour = build_if_needed()
    return 'OK (end_hour=%s)' % end_hour



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Class for helping manage a crash result."""

# pylint: disable=unpacking-non-sequence

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.crash_analysis import crash_analyzer
from clusterfuzz._internal.crash_analysis.stack_parsing import stack_analyzer


class CrashResult(object):
  """Represents a crash result from a test run."""

  def __init__(self, return_code, crash_time, output):
    self.return_code = return_code
    self.crash_time = crash_time
    self.output = utils.decode_to_unicode(output) if output else 'No output!'

    self._symbolized_crash_data = None
    self._unsymbolized_crash_data = None

  def get_crash_time(self):
    """Return the crash time."""
    return self.crash_time

  def get_symbolized_data(self):
    """Compute symbolized crash data if necessary or return cached result."""
    if self._symbolized_crash_data:
      return self._symbolized_crash_data

    self._symbolized_crash_data = stack_analyzer.get_crash_data(
        self.output, symbolize_flag=True)
    return self._symbolized_crash_data

  def get_unsymbolized_data(self):
    """Compute unsymbolized crash data if necessary or return cached result."""
    if self._unsymbolized_crash_data:
      return self._unsymbolized_crash_data

    self._unsymbolized_crash_data = stack_analyzer.get_crash_data(
        self.output, symbolize_flag=False)
    return self._unsymbolized_crash_data

  def get_state(self, symbolized=True):
    """Return the crash state."""
    if symbolized:
      state = self.get_symbolized_data()
    else:
      state = self.get_unsymbolized_data()

    return state.crash_state

  def get_stacktrace(self, symbolized=True):
    """Return the crash stacktrace."""
    if symbolized:
      state = self.get_symbolized_data()
    else:
      state = self.get_unsymbolized_data()

    return state.crash_stacktrace

  def get_type(self):
    """Return the crash type."""
    # It does not matter whether we use symbolized or unsymbolized data.
    state = self.get_unsymbolized_data()
    return state.crash_type

  def is_crash(self, ignore_state=False):
    """Return True if this result was a crash."""
    crashed = crash_analyzer.is_crash(self.return_code, self.output)
    if not crashed:
      return False

    state = self.get_state(symbolized=False)
    if not state.strip() and not ignore_state:
      return False
    return True

  def should_ignore(self):
    """Return True if this crash should be ignored."""
    state = self.get_symbolized_data()
    return crash_analyzer.ignore_stacktrace(state.crash_stacktrace)

  def is_security_issue(self):
    """Return True if this crash is a security issue."""
    state = self.get_unsymbolized_data()
    return crash_analyzer.is_security_issue(
        state.crash_stacktrace, state.crash_type, state.crash_address)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Query handles pagination and OR conditions with its best effort."""

from google.cloud.ndb import exceptions

from libs.query import base


def _get_key_fn(attribute_name):
  """Return the function to get attr of an item. This is used in sorting."""

  def get_key(item):
    return getattr(item, attribute_name)

  return get_key


def compute_projection(projection, order_property):
  """Set projection."""
  if projection is None:
    return None

  combined_projection = set(projection)
  combined_projection.add(order_property)
  return list(combined_projection)


def _combine(q1, q2):
  """Combine KeyQuery q1 and q2. We ignore or_filters because we assume q1 and
    q2 are flat. In other words, they are results of _KeyQuery.flatten(..)."""
  assert not q1.or_filters
  assert not q2.or_filters
  assert q1.order_property == q2.order_property
  assert q1.order_desc == q2.order_desc

  result = _KeyQuery(q1.model)
  result.filters = q1.filters + q2.filters
  result.order_property = q1.order_property
  result.order_desc = q1.order_desc
  return result


class _Run(object):
  """Encapsulate a query and its run."""

  def __init__(self, query, **kwargs):
    self.query = query
    self.result = query.iter(**kwargs)


class _KeyQuery(object):
  """Query only keys. It supports an OR condition."""

  def __init__(self, model):
    self.model = model
    self.or_filters = []
    self.filters = []
    self.order_property = None
    self.order_desc = False

  def union(self, *queries):
    """Specify the OR condition."""
    self.or_filters.append(queries)

  def filter(self, operator, prop, value):
    """Specify the filter."""
    if operator == 'IN':
      subqueries = []
      for v in value:
        q = _KeyQuery(self.model)
        q.filter('=', prop, v)
        subqueries.append(q)
      self.union(*subqueries)
    else:
      self.filters.append((operator, prop, value))

  def order(self, prop, is_desc):
    """Specify the order."""
    self.order_property, self.order_desc = prop, is_desc

  def flatten(self):
    """Flatten self into multiple queries if or_filters is not empty."""
    if not self.or_filters:
      return [self]

    for qs in self.or_filters:
      for q in qs:
        q.order(self.order_property, self.order_desc)

    queries = []
    for query in self.or_filters[0]:
      for q in query.flatten():
        queries.append(q)

    for or_queries in self.or_filters[1:]:
      new_queries = []
      for oq in or_queries:
        for fq in oq.flatten():
          for q in queries:
            new_queries.append(_combine(q, fq))
      queries = new_queries

    for q in queries:
      for (prop_op, prop, value) in self.filters:
        q.filter(prop_op, prop, value)

    return queries

  def to_datastore_query(self):
    """Return the corresponding datastore query."""
    assert not self.or_filters

    query = self.model.query()
    properties = self.model._properties  # pylint: disable=protected-access
    for (prop_op, prop, value) in self.filters:
      if prop_op == '=':
        filter_func = properties[prop].__eq__
      elif prop_op == '!=':
        filter_func = properties[prop].__ne__
      elif prop_op == '<':
        filter_func = properties[prop].__le__
      elif prop_op == '>':
        filter_func = properties[prop].__gt__
      elif prop_op == '<=':
        filter_func = properties[prop].__le__
      elif prop_op == '>=':
        filter_func = properties[prop].__ge__

      query = query.filter(filter_func(value))

    if self.order_property:
      order_property = properties[self.order_property]
      if self.order_desc:
        order_property = -order_property

      query = query.order(order_property)

    return query

  def _build_runs(self, total):
    """Construct queries and run them."""
    queries = self.flatten()

    runs = []
    # TODO(tanin): Improve the speed by detecting if we need union (or OR).
    # If we don't need union, we can set keys_only=True and projection=None in
    # order to improve speed; it's likely to be 2x faster.
    for q in queries:
      runs.append(
          _Run(
              q.to_datastore_query(),
              keys_only=False,
              projection=[self.order_property],
              limit=total))
    return runs

  def _get_total_count(self, runs, offset, limit, items, more_limit):
    """Get total count by querying more items."""
    max_total_count = offset + limit + more_limit
    current_count = len(items)

    if current_count > max_total_count:
      return max_total_count, True

    more_limit += 1
    more_runs = []
    for run in runs:
      try:
        cursor = run.result.cursor_after()
      except exceptions.BadArgumentError:
        # iterator had no results.
        cursor = None

      more_runs.append(
          _Run(
              run.query,
              start_cursor=cursor,
              keys_only=True,
              projection=None,
              limit=more_limit))

    keys = {item.key.id() for item in items}
    for run in more_runs:
      for key in run.result:
        keys.add(key)

    total_count = min(len(keys), max_total_count)
    has_more = len(keys) >= max_total_count

    return total_count, has_more

  def fetch(self, offset, limit, more_limit):
    """Construct multiple queries based on the or_filters, query them,
      combined the results, return items and total_count."""
    runs = self._build_runs(offset + limit)

    items = {}
    for run in runs:
      for item in run.result:
        if item.key.id() not in items:
          items[item.key.id()] = item

    items = sorted(
        list(items.values()),
        reverse=self.order_desc,
        key=_get_key_fn(self.order_property))

    total_count, has_more = self._get_total_count(runs, offset, limit, items,
                                                  more_limit)

    return items[offset:(offset + limit)], total_count, has_more


class Query(base.Query):
  """Query that returns items with a smart count. A smart count indicates
    a lowerbound of the total count (e.g. 500+)."""

  def __init__(self, model):
    self.model = model
    self.key_query = _KeyQuery(model)
    self.order_property = None
    self.order_desc = False

  def filter(self, field, value, operator='='):
    """Specify the filter."""
    self.key_query.filter(operator, field, value)

  def filter_in(self, field, values):
    """Specify the filter IN."""
    self.key_query.filter('IN', field, values)

  def union(self, *queries):
    """Specify the OR condition."""
    self.key_query.union(*[q.key_query for q in queries])

  def new_subquery(self):
    """Generate a new subquery with the same model."""
    return Query(self.model)

  def order(self, field, is_desc):
    """Specify the order."""
    self.order_property = field
    self.order_desc = is_desc
    self.key_query.order(field, is_desc)

  def fetch(self, offset, limit, projection, more_limit):
    """Return the items, the total count, and if there are more number of items
      than the total count."""
    assert self.order_property

    keys, total_count, has_more = self.key_query.fetch(
        limit=limit, offset=offset, more_limit=more_limit)

    if keys:
      item_query = self.key_query.model.query(
          self.key_query.model.key.IN([key.key for key in keys]))

      items = item_query.fetch(
          limit=limit,
          projection=compute_projection(projection, self.order_property))
      items = sorted(
          items, reverse=self.order_desc, key=_get_key_fn(self.order_property))
    else:
      items = []

    return items, total_count, has_more

  def fetch_page(self, page, page_size, projection, more_limit):
    """Return the items, total_pages, total_items, and has_more."""

    # Validation check to convert all negative page numbers to 1.
    if page < 1:
      page = 1

    items, total_items, has_more = self.fetch(
        offset=(page - 1) * page_size,
        limit=page_size,
        projection=projection,
        more_limit=more_limit)

    total_pages = total_items // page_size
    if (total_items % page_size) > 0:
      total_pages += 1

    return items, total_pages, total_items, has_more



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""run_bot.py run a Clusterfuzz bot locally."""
import os
import signal

from local.butler import appengine
from local.butler import common
from local.butler import constants


def _setup_bot_directory(args):
  """Set up the bot directory."""
  appengine.symlink_config_dir()

  src_root_dir = os.path.abspath('.')
  if os.path.exists(args.directory):
    print('Bot directory already exists. Re-using...')
  else:
    print('Creating new CF bot directory...')
    os.makedirs(args.directory)

  clusterfuzz_dir = os.path.join(args.directory, 'clusterfuzz')
  bot_src_dir = os.path.join(clusterfuzz_dir, 'src')
  if not os.path.exists(clusterfuzz_dir):
    os.makedirs(clusterfuzz_dir)
    os.mkdir(bot_src_dir)

  common.update_dir(
      os.path.join(src_root_dir, 'src', 'appengine'),
      os.path.join(bot_src_dir, 'appengine'))
  common.update_dir(
      os.path.join(src_root_dir, 'src', 'python'),
      os.path.join(bot_src_dir, 'python'))
  common.update_dir(
      os.path.join(src_root_dir, 'src', 'clusterfuzz'),
      os.path.join(bot_src_dir, 'clusterfuzz'))
  common.update_dir(
      os.path.join(src_root_dir, 'src', 'third_party'),
      os.path.join(bot_src_dir, 'third_party'))

  common.update_dir(
      os.path.join(src_root_dir, 'resources'),
      os.path.join(clusterfuzz_dir, 'resources'))
  common.update_dir(
      os.path.join(src_root_dir, 'bot'), os.path.join(clusterfuzz_dir, 'bot'))


def _setup_environment_and_configs(args, appengine_path):
  """Set up environment variables and configuration files."""
  clusterfuzz_dir = os.path.abspath(os.path.join(args.directory, 'clusterfuzz'))

  # Matches startup scripts.
  os.environ['PYTHONPATH'] = ':'.join([
      os.getenv('PYTHONPATH', ''),
      appengine_path,
      os.path.join(clusterfuzz_dir, 'src'),
  ])

  os.environ['ROOT_DIR'] = clusterfuzz_dir
  if not os.getenv('BOT_NAME'):
    os.environ['BOT_NAME'] = args.name

  os.environ['LD_LIBRARY_PATH'] = '{0}:{1}'.format(
      os.path.join(clusterfuzz_dir, 'src', 'clusterfuzz', '_internal',
                   'scripts'), os.getenv('LD_LIBRARY_PATH', ''))

  tmpdir = os.path.join(clusterfuzz_dir, 'bot_tmpdir')
  if not os.path.exists(tmpdir):
    os.mkdir(tmpdir)
  os.environ['TMPDIR'] = tmpdir
  os.environ['BOT_TMPDIR'] = tmpdir

  os.environ['KILL_STALE_INSTANCES'] = 'False'
  os.environ['LOCAL_DEVELOPMENT'] = 'True'
  os.environ['DATASTORE_EMULATOR_HOST'] = constants.DATASTORE_EMULATOR_HOST
  os.environ['PUBSUB_EMULATOR_HOST'] = constants.PUBSUB_EMULATOR_HOST
  os.environ['APPLICATION_ID'] = constants.TEST_APP_ID

  if not os.getenv('UNTRUSTED_WORKER'):
    local_gcs_buckets_path = os.path.abspath(
        os.path.join(args.server_storage_path, 'local_gcs'))
    assert os.path.exists(local_gcs_buckets_path), (
        'Server storage path not found, make sure to start run_server with '
        'the same storage path.')

    os.environ['LOCAL_GCS_BUCKETS_PATH'] = local_gcs_buckets_path

  if args.android_serial:
    if not os.getenv('OS_OVERRIDE'):
      os.environ['OS_OVERRIDE'] = 'ANDROID'

    os.environ['ANDROID_SERIAL'] = args.android_serial


def execute(args):
  """Run the bot."""
  appengine_path = appengine.find_sdk_path()

  _setup_bot_directory(args)
  _setup_environment_and_configs(args, appengine_path)

  try:
    os.chdir(os.path.join(args.directory, 'clusterfuzz'))
    proc = common.execute_async('python src/python/bot/startup/run_bot.py')

    def _stop_handler(*_):
      print('Bot has been stopped. Exit.')
      proc.kill()

    signal.signal(signal.SIGTERM, _stop_handler)
    common.process_proc_output(proc)
    proc.wait()
  except KeyboardInterrupt:
    _stop_handler()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""server.py initialises the appengine server for ClusterFuzz."""

from flask import Flask
from flask import redirect
from flask import request
from google.cloud import ndb

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment
from handlers import base_handler
from handlers import bots
from handlers import commit_range
from handlers import configuration
from handlers import corpora
from handlers import coverage_report
from handlers import crash_query
from handlers import crash_stats
from handlers import domain_verifier
from handlers import download
from handlers import external_update
from handlers import fuzzer_stats
from handlers import fuzzers
from handlers import gcs_redirector
from handlers import help_redirector
from handlers import home
from handlers import issue_redirector
from handlers import jobs
from handlers import login
from handlers import report_csp_failure
from handlers import revisions_info
from handlers import testcase_list
from handlers import upload_testcase
from handlers import viewer
from handlers.cron import backup
from handlers.cron import batch_fuzzer_jobs
from handlers.cron import build_crash_stats
from handlers.cron import cleanup
from handlers.cron import corpus_backup
from handlers.cron import fuzz_strategy_selection
from handlers.cron import fuzzer_and_job_weights
from handlers.cron import fuzzer_coverage
from handlers.cron import load_bigquery_stats
from handlers.cron import manage_vms
from handlers.cron import ml_train
from handlers.cron import oss_fuzz_apply_ccs
from handlers.cron import oss_fuzz_build_status
from handlers.cron import oss_fuzz_generate_certs
from handlers.cron import predator_pull
from handlers.cron import project_setup
from handlers.cron import recurring_tasks
from handlers.cron import schedule_corpus_pruning
from handlers.cron import sync_admins
from handlers.cron import triage
from handlers.performance_report import show as show_performance_report
from handlers.reproduce_tool import get_config
from handlers.reproduce_tool import testcase_info
from handlers.testcase_detail import crash_stats as crash_stats_on_testcase
from handlers.testcase_detail import create_issue
from handlers.testcase_detail import delete
from handlers.testcase_detail import download_testcase
from handlers.testcase_detail import find_similar_issues
from handlers.testcase_detail import mark_fixed
from handlers.testcase_detail import mark_security
from handlers.testcase_detail import mark_unconfirmed
from handlers.testcase_detail import redo
from handlers.testcase_detail import remove_duplicate
from handlers.testcase_detail import remove_group
from handlers.testcase_detail import remove_issue
from handlers.testcase_detail import show as show_testcase
from handlers.testcase_detail import testcase_variants
from handlers.testcase_detail import update_from_trunk
from handlers.testcase_detail import update_issue

ndb_client = ndb.Client()


def ndb_wsgi_middleware(wsgi_app):
  """WSGI middleware for ndb_datastore context allocation to the app."""

  def middleware(environ, start_response):
    with ndb_client.context():
      return wsgi_app(environ, start_response)

  return middleware


def register_routes(flask_app, routes):
  """Utility function to register all routes to the flask app."""
  for route, handler in routes:
    flask_app.add_url_rule(route, view_func=handler.as_view(route))


# Add item to the navigation menu. Order is important.
base_handler.add_menu('Testcases', '/testcases')
base_handler.add_menu('Fuzzer Statistics', '/fuzzer-stats')
base_handler.add_menu('Crash Statistics', '/crash-stats')
base_handler.add_menu('Upload Testcase', '/upload-testcase')

_is_chromium = utils.is_chromium()
_is_oss_fuzz = utils.is_oss_fuzz()

if _is_chromium:
  base_handler.add_menu('Crashes by range', '/commit-range')

if not _is_oss_fuzz:
  base_handler.add_menu('Fuzzers', '/fuzzers')
  base_handler.add_menu('Corpora', '/corpora')
  base_handler.add_menu('Bots', '/bots')

base_handler.add_menu('Jobs', '/jobs')
base_handler.add_menu('Configuration', '/configuration')
base_handler.add_menu('Report Bug', '/report-bug')
base_handler.add_menu('Documentation', '/docs')

# We need to separate routes for cron to avoid redirection.
cron_routes = [
    ('/backup', backup.Handler),
    ('/batch-fuzzer-jobs', batch_fuzzer_jobs.Handler),
    ('/build-crash-stats', build_crash_stats.Handler),
    ('/cleanup', cleanup.Handler),
    ('/corpus-backup/make-public', corpus_backup.MakePublicHandler),
    ('/external-update', external_update.Handler),
    ('/fuzzer-and-job-weights', fuzzer_and_job_weights.Handler),
    ('/fuzzer-coverage', fuzzer_coverage.Handler),
    ('/fuzzer-stats/cache', fuzzer_stats.RefreshCacheHandler),
    ('/fuzzer-stats/preload', fuzzer_stats.PreloadHandler),
    ('/fuzz-strategy-selection', fuzz_strategy_selection.Handler),
    ('/home-cache', home.RefreshCacheHandler),
    ('/load-bigquery-stats', load_bigquery_stats.Handler),
    ('/manage-vms', manage_vms.Handler),
    ('/oss-fuzz-apply-ccs', oss_fuzz_apply_ccs.Handler),
    ('/oss-fuzz-build-status', oss_fuzz_build_status.Handler),
    ('/oss-fuzz-generate-certs', oss_fuzz_generate_certs.Handler),
    ('/project-setup', project_setup.Handler),
    ('/predator-pull', predator_pull.Handler),
    ('/schedule-corpus-pruning', schedule_corpus_pruning.Handler),
    ('/schedule-impact-tasks', recurring_tasks.ImpactTasksScheduler),
    ('/schedule-ml-train-tasks', ml_train.Handler),
    ('/schedule-progression-tasks', recurring_tasks.ProgressionTasksScheduler),
    ('/schedule-upload-reports-tasks',
     recurring_tasks.UploadReportsTaskScheduler),
    ('/sync-admins', sync_admins.Handler),
    ('/testcases/cache', testcase_list.CacheHandler),
    ('/triage', triage.Handler),
]

handlers = [
    ('/', home.Handler if _is_oss_fuzz else testcase_list.Handler),
    ('/add-external-user-permission', configuration.AddExternalUserPermission),
    ('/bots', bots.Handler),
    ('/bots/dead', bots.DeadBotsHandler),
    ('/bots/load', bots.JsonHandler),
    ('/commit-range', commit_range.Handler),
    ('/commit-range/load', commit_range.JsonHandler),
    ('/configuration', configuration.Handler),
    ('/coverage-report', coverage_report.Handler),
    ('/coverage-report/<report_type>/<argument>/<date>',
     coverage_report.Handler),
    ('/coverage-report/<report_type>/<argument>/<date>/<path:extra>',
     coverage_report.Handler),
    ('/crash-query', crash_query.Handler),
    ('/crash-stats/load', crash_stats.JsonHandler),
    ('/crash-stats', crash_stats.Handler),
    ('/corpora', corpora.Handler),
    ('/corpora/create', corpora.CreateHandler),
    ('/corpora/delete', corpora.DeleteHandler),
    ('/delete-external-user-permission',
     configuration.DeleteExternalUserPermission),
    ('/docs', help_redirector.DocumentationHandler),
    ('/download', download.Handler),
    ('/download/<resource>', download.Handler),
    ('/fuzzer-stats/load', fuzzer_stats.LoadHandler),
    ('/fuzzer-stats/load-filters', fuzzer_stats.LoadFiltersHandler),
    ('/fuzzer-stats', fuzzer_stats.Handler),
    ('/fuzzer-stats/<path:extra>', fuzzer_stats.Handler),
    ('/fuzzers', fuzzers.Handler),
    ('/fuzzers/create', fuzzers.CreateHandler),
    ('/fuzzers/delete', fuzzers.DeleteHandler),
    ('/fuzzers/edit', fuzzers.EditHandler),
    ('/fuzzers/log/<fuzzer_name>', fuzzers.LogHandler),
    ('/gcs-redirect', gcs_redirector.Handler),
    ('/google<tag>.html', domain_verifier.Handler),
    ('/issue', issue_redirector.Handler),
    ('/issue/<testcase_id>', issue_redirector.Handler),
    ('/jobs', jobs.Handler),
    ('/jobs/load', jobs.JsonHandler),
    ('/jobs/delete-job', jobs.DeleteJobHandler),
    ('/login', login.Handler),
    ('/logout', login.LogoutHandler),
    ('/performance-report', show_performance_report.Handler),
    ('/performance-report/<fuzzer_name>/<job_type>/<logs_date>',
     show_performance_report.Handler),
    ('/reproduce-tool/get-config', get_config.Handler),
    ('/reproduce-tool/testcase-info', testcase_info.Handler),
    ('/report-bug', help_redirector.ReportBugHandler),
    ('/report-csp-failure', report_csp_failure.ReportCspFailureHandler),
    ('/revisions', revisions_info.Handler),
    ('/session-login', login.SessionLoginHandler),
    ('/testcase', show_testcase.DeprecatedHandler),
    ('/testcase-detail', show_testcase.Handler),
    ('/testcase-detail/<int:testcase_id>', show_testcase.Handler),
    ('/testcase-detail/crash-stats', crash_stats_on_testcase.Handler),
    ('/testcase-detail/create-issue', create_issue.Handler),
    ('/testcase-detail/delete', delete.Handler),
    ('/testcase-detail/download-testcase', download_testcase.Handler),
    ('/testcase-detail/find-similar-issues', find_similar_issues.Handler),
    ('/testcase-detail/mark-fixed', mark_fixed.Handler),
    ('/testcase-detail/mark-security', mark_security.Handler),
    ('/testcase-detail/mark-unconfirmed', mark_unconfirmed.Handler),
    ('/testcase-detail/redo', redo.Handler),
    ('/testcase-detail/refresh', show_testcase.RefreshHandler),
    ('/testcase-detail/remove-duplicate', remove_duplicate.Handler),
    ('/testcase-detail/remove-issue', remove_issue.Handler),
    ('/testcase-detail/remove-group', remove_group.Handler),
    ('/testcase-detail/testcase-variants', testcase_variants.Handler),
    ('/testcase-detail/update-from-trunk', update_from_trunk.Handler),
    ('/testcase-detail/update-issue', update_issue.Handler),
    ('/testcases', testcase_list.Handler),
    ('/testcases/load', testcase_list.JsonHandler),
    ('/upload-testcase', upload_testcase.Handler),
    ('/upload-testcase/get-url-oauth', upload_testcase.UploadUrlHandlerOAuth),
    ('/upload-testcase/prepare', upload_testcase.PrepareUploadHandler),
    ('/upload-testcase/load', upload_testcase.JsonHandler),
    ('/upload-testcase/upload', upload_testcase.UploadHandler),
    ('/upload-testcase/upload-oauth', upload_testcase.UploadHandlerOAuth),
    ('/update-job', jobs.UpdateJob),
    ('/update-job-template', jobs.UpdateJobTemplate),
    ('/viewer', viewer.Handler),
]

logs.configure('appengine')
config = local_config.GAEConfig()
main_domain = config.get('domains.main')
redirect_domains = config.get('domains.redirects')


def redirect_handler():
  """Redirection handler."""
  if not redirect_domains:
    return None

  if request.host in redirect_domains:
    return redirect('https://' + main_domain + request.full_path)

  return None


app = Flask(__name__)
# To also process trailing slash urls.
app.url_map.strict_slashes = False
app.before_request(redirect_handler)

if not environment.get_value('PY_UNITTESTS'):
  # Adding ndb context middleware when not running tests.
  app.wsgi_app = ndb_wsgi_middleware(app.wsgi_app)

register_routes(app, handlers)
register_routes(app, cron_routes)

if __name__ == '__main__':
  app.run(host=main_domain)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuzzing engine interface."""

_ENGINES = {}


class Error(Exception):
  """Engine error."""


class FuzzOptions(object):
  """Represents options passed to the engine. Can be overridden to provide more
  options."""

  def __init__(self, corpus_dir, arguments, strategies):
    self.corpus_dir = corpus_dir
    self.arguments = arguments
    self.strategies = strategies


class Crash(object):
  """Represents a crash found by the fuzzing engine."""

  def __init__(self, input_path, stacktrace, reproduce_args, crash_time):
    self.input_path = input_path
    self.stacktrace = stacktrace
    self.reproduce_args = reproduce_args
    self.crash_time = crash_time


class FuzzResult(object):
  """Represents a result of a fuzzing session: a list of crashes found and the
  stats generated."""

  def __init__(self, logs, command, crashes, stats, time_executed):
    self.logs = logs
    self.command = command
    self.crashes = crashes
    self.stats = stats
    self.time_executed = time_executed


class ReproduceResult(object):
  """Results from running a testcase against a target."""

  def __init__(self, command, return_code, time_executed, output):
    self.command = command
    self.return_code = return_code
    self.time_executed = time_executed
    self.output = output


class Engine(object):
  """Base interface for a grey box fuzzing engine."""

  def __init__(self):
    self.do_strategies = True

  @property
  def name(self):
    """Get the name of the engine."""
    raise NotImplementedError

  def fuzz_additional_processing_timeout(self, options):
    """Return the maximum additional timeout in seconds for additional
    operations in fuzz() (e.g. merging back new items).

    Args:
      options: A FuzzOptions object.

    Returns:
      An int representing the number of seconds required.
    """
    del options
    return 0

  def prepare(self, corpus_dir, target_path, build_dir):
    """Prepare for a fuzzing session, by generating options. Returns a
    FuzzOptions object.

    Args:
      corpus_dir: The main corpus directory.
      target_path: Path to the target.
      build_dir: Path to the build directory.

    Returns:
      A FuzzOptions object.
    """
    raise NotImplementedError

  def fuzz(self, target_path, options, reproducers_dir, max_time):
    """Run a fuzz session.

    Args:
      target_path: Path to the target.
      options: The FuzzOptions object returned by prepare().
      reproducers_dir: The directory to put reproducers in when crashes
          are found.
      max_time: Maximum allowed time for the fuzzing to run.

    Returns:
      A FuzzResult object.
    """
    raise NotImplementedError

  def reproduce(self, target_path, input_path, arguments, max_time):
    """Reproduce a crash given an input.

    Args:
      target_path: Path to the target.
      input_path: Path to the reproducer input.
      arguments: Additional arguments needed for reproduction.
      max_time: Maximum allowed time for the reproduction.

    Returns:
      A ReproduceResult.

    Raises:
      TimeoutError: If the reproduction exceeds max_time.
    """
    raise NotImplementedError

  def minimize_corpus(self, target_path, arguments, input_dirs, output_dir,
                      reproducers_dir, max_time):
    """Optional (but recommended): run corpus minimization.

    Args:
      target_path: Path to the target.
      arguments: Additional arguments needed for corpus minimization.
      input_dirs: Input corpora.
      output_dir: Output directory to place minimized corpus.
      reproducers_dir: The directory to put reproducers in when crashes are
          found.
      max_time: Maximum allowed time for the minimization.

    Returns:
      A FuzzResult object.

    Raises:
      TimeoutError: If the corpus minimization exceeds max_time.
      Error: If the merge failed in some other way.
    """
    raise NotImplementedError

  def minimize_testcase(self, target_path, arguments, input_path, output_path,
                        max_time):
    """Optional (but recommended): Minimize a testcase.

    Args:
      target_path: Path to the target.
      arguments: Additional arguments needed for testcase minimization.
      input_path: Path to the reproducer input.
      output_path: Path to the minimized output.
      max_time: Maximum allowed time for the minimization.

    Returns:
      A ReproduceResult.

    Raises:
      TimeoutError: If the testcase minimization exceeds max_time.
    """
    raise NotImplementedError

  def cleanse(self, target_path, arguments, input_path, output_path, max_time):
    """Optional (but recommended): Cleanse a testcase.

    Args:
      target_path: Path to the target.
      arguments: Additional arguments needed for testcase cleanse.
      input_path: Path to the reproducer input.
      output_path: Path to the cleansed output.
      max_time: Maximum allowed time for the cleanse.

    Returns:
      A ReproduceResult.

    Raises:
      TimeoutError: If the cleanse exceeds max_time.
    """
    raise NotImplementedError


def register(name, engine_class):
  """Register a fuzzing engine."""
  if name in _ENGINES:
    raise ValueError('Engine {name} is already registered'.format(name=name))

  _ENGINES[name] = engine_class


def get(name):
  """Get an implemntation of a fuzzing engine, or None if one does not exist."""
  engine_class = _ENGINES.get(name)
  if engine_class:
    return engine_class()

  return None



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Process results from Predator and update test cases accordingly."""

import json

from clusterfuzz._internal.base import errors
from clusterfuzz._internal.config import db_config
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.google_cloud_utils import pubsub
from clusterfuzz._internal.metrics import logs
from handlers import base_handler
from libs import handler


class Handler(base_handler.Handler):
  """Handler to periodically gather new results from Predator requests."""

  @handler.cron()
  def get(self):
    """Process a GET request."""
    subscription = db_config.get_value('predator_result_topic')
    if not subscription:
      logs.log('No Predator subscription configured. Aborting.')
      return

    client = pubsub.PubSubClient()
    messages = client.pull_from_subscription(subscription, acknowledge=True)
    for message in messages:
      message = json.loads(message.data)
      testcase_id = message['crash_identifiers']
      try:
        testcase = data_handler.get_testcase_by_id(testcase_id)
      except errors.InvalidTestcaseError:
        logs.log('Testcase %s no longer exists.' % str(testcase_id))
        continue

      testcase.set_metadata('predator_result', message, update_testcase=False)
      testcase.delete_metadata('blame_pending', update_testcase=False)
      testcase.put()
      logs.log('Set predator result for testcase %d.' % testcase.key.id())

    logs.log('Finished processing predator results. %d total.' % len(messages))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""access tests."""
import unittest

# pylint: disable=protected-access
import mock

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import mock_config
from clusterfuzz._internal.tests.test_libs import test_utils
from libs import access
from libs import auth
from libs import helpers
from libs.issue_management import monorail
from libs.issue_management.monorail import issue
from libs.issue_management.monorail import issue_tracker_manager


class GetUserJobTypeTest(unittest.TestCase):
  """Test get_user_job_type."""

  _FAKE_CONFIG = 'test@test.com\nTest2@test.com;job'

  def setUp(self):
    test_helpers.patch(self, [
        'clusterfuzz._internal.config.db_config.get_value',
        'libs.helpers.get_user_email'
    ])

  def test_none(self):
    """Ensure it returns None when config is invalid."""
    self.mock.get_value.return_value = None
    self.mock.get_user_email.return_value = ''
    self.assertIsNone(access.get_user_job_type())
    self.mock.get_value.assert_has_calls([mock.call('privileged_users')])

  def test_get_job(self):
    """Test getting job for a specific user."""
    self.mock.get_value.return_value = self._FAKE_CONFIG
    self.mock.get_user_email.return_value = 'test2@test.com'
    self.assertEqual('job', access.get_user_job_type())

  def test_get_none(self):
    """Test getting None when the user is globally privileged."""
    self.mock.get_value.return_value = self._FAKE_CONFIG
    self.mock.get_user_email.return_value = 'test@test.com'
    self.assertIsNone(access.get_user_job_type())


class IsPrivilegedUserTest(unittest.TestCase):
  """Test _is_privileged_user."""

  _FAKE_CONFIG = 'Test@test.com\n'

  def setUp(self):
    test_helpers.patch(self,
                       ['clusterfuzz._internal.config.db_config.get_value'])

  def test_none(self):
    """Ensure it returns False when config is invalid."""
    self.mock.get_value.return_value = None
    self.assertFalse(access._is_privileged_user('test'))
    self.mock.get_value.assert_has_calls([mock.call('privileged_users')])

  def test_global(self):
    """Ensure it returns True if an email is permitted globally."""
    self.mock.get_value.return_value = self._FAKE_CONFIG
    self.assertTrue(access._is_privileged_user('test@test.com'))
    self.mock.get_value.assert_has_calls([mock.call('privileged_users')])

  def test_all_users_privileged(self):
    """Test when all_users_privileged is set."""
    test_helpers.patch(self,
                       ['clusterfuzz._internal.config.local_config.AuthConfig'])
    self.mock.AuthConfig.return_value = mock_config.MockConfig({
        'all_users_privileged': True,
    })
    self.assertTrue(access._is_privileged_user('a@user.com'))


class IsDomainAllowedTest(unittest.TestCase):
  """Test _is_domain_allowed."""

  _FAKE_CONFIG = ['test.com', 'test2.com']

  def setUp(self):
    test_helpers.patch(
        self, ['clusterfuzz._internal.config.local_config.AuthConfig.get'])

  def test_none(self):
    """Ensure it returns False when config is invalid."""
    self.mock.get.return_value = []
    self.assertFalse(access._is_domain_allowed('test@test.com'))

  def test_allow(self):
    """Ensure it returns True when the domain is on the list."""
    self.mock.get.return_value = self._FAKE_CONFIG
    self.assertTrue(access._is_domain_allowed('test@test.com'))
    self.assertTrue(access._is_domain_allowed('test@test2.com'))

  def test_forbid(self):
    """Ensure it returns False when the domain is not on the list."""
    self.mock.get.return_value = self._FAKE_CONFIG
    self.assertFalse(access._is_domain_allowed('test.com'))
    self.assertFalse(access._is_domain_allowed('test@test24234.com'))


class GetAccessTest(unittest.TestCase):
  """Test get_access."""

  def setUp(self):
    test_helpers.patch(self, [
        'libs.auth.get_current_user',
        'libs.auth.is_current_user_admin',
        'libs.access._is_blacklisted_user',
        'libs.access._is_privileged_user',
        'libs.access._is_domain_allowed',
        'clusterfuzz._internal.base.external_users.is_fuzzer_allowed_for_user',
        'clusterfuzz._internal.base.external_users.is_job_allowed_for_user',
    ])
    self.mock._is_blacklisted_user.return_value = False
    self.user = auth.User('test@test.com')

  def test_get_access_access_redirect(self):
    """Ensure it redirects when user is None."""
    self.mock.get_current_user.return_value = None
    self.mock.is_current_user_admin.return_value = False
    self.assertEqual(access.get_access(), access.UserAccess.Redirected)

  def test_get_access_admin(self):
    """Ensure it allows when user is admin."""
    self.mock.get_current_user.return_value = self.user
    self.mock.is_current_user_admin.return_value = True
    self.assertEqual(access.get_access(), access.UserAccess.Allowed)

  def test_get_access_privileged(self):
    """Ensure it allows when user is privileged."""
    self.mock.get_current_user.return_value = self.user
    self.mock.is_current_user_admin.return_value = False
    self.mock._is_privileged_user.return_value = True
    self.assertEqual(access.get_access(), access.UserAccess.Allowed)

  def test_get_access_allowed_domain(self):
    """For an allowed domain, ensure it allows when privileged is not needed,
       and ensure it denies when privileged is needed."""
    self.mock.get_current_user.return_value = self.user
    self.mock.is_current_user_admin.return_value = False
    self.mock._is_privileged_user.return_value = False
    self.mock._is_domain_allowed.return_value = True
    self.assertEqual(
        access.get_access(need_privileged_access=False),
        access.UserAccess.Allowed)
    self.assertEqual(
        access.get_access(need_privileged_access=True),
        access.UserAccess.Denied)

  def test_get_access_external_fuzzer(self):
    """For a fuzzer, ensure it allows when a user is allowed."""
    self.mock.get_current_user.return_value = self.user
    self.mock.is_current_user_admin.return_value = False
    self.mock._is_privileged_user.return_value = False
    self.mock._is_domain_allowed.return_value = False
    self.mock.is_fuzzer_allowed_for_user.return_value = True
    self.mock.is_job_allowed_for_user.return_value = False
    self.assertEqual(
        access.get_access(fuzzer_name='test'), access.UserAccess.Allowed)
    self.assertEqual(access.get_access(), access.UserAccess.Denied)

  def test_get_access_external_job(self):
    """For a job, ensure it allows when a user is allowed."""
    self.mock.get_current_user.return_value = self.user
    self.mock.is_current_user_admin.return_value = False
    self.mock._is_privileged_user.return_value = False
    self.mock._is_domain_allowed.return_value = False
    self.mock.is_fuzzer_allowed_for_user.return_value = False
    self.mock.is_job_allowed_for_user.return_value = True
    self.assertEqual(
        access.get_access(job_type='test'), access.UserAccess.Allowed)
    self.assertEqual(access.get_access(), access.UserAccess.Denied)

  def test_get_access_denied(self):
    """Ensure it denies when every condition is false."""
    self.mock.get_current_user.return_value = self.user
    self.mock.is_current_user_admin.return_value = False
    self.mock._is_privileged_user.return_value = False
    self.mock._is_domain_allowed.return_value = False
    self.mock.is_fuzzer_allowed_for_user.return_value = False
    self.mock.is_job_allowed_for_user.return_value = False
    self.assertEqual(
        access.get_access(fuzzer_name='test'), access.UserAccess.Denied)
    self.assertEqual(access.get_access(), access.UserAccess.Denied)

  def test_blacklisted_user(self):
    """Test blacklisted users."""
    self.mock.is_current_user_admin.return_value = False
    self.mock._is_domain_allowed.return_value = True
    self.mock._is_blacklisted_user.return_value = True
    self.assertEqual(access.get_access(), access.UserAccess.Denied)


class HasAccessTest(unittest.TestCase):
  """Test has_access."""

  def setUp(self):
    test_helpers.patch(self, [
        'libs.access.get_access',
    ])

  def test_allowed(self):
    """Test allowed."""
    self.mock.get_access.return_value = access.UserAccess.Allowed
    self.assertTrue(access.has_access(True, 'a', 'b'))
    self.mock.get_access.assert_has_calls([mock.call(True, 'a', 'b')])

  def test_denied(self):
    """Test denied."""
    self.mock.get_access.return_value = access.UserAccess.Denied
    self.assertFalse(access.has_access(True, 'a', 'b'))
    self.mock.get_access.assert_has_calls([mock.call(True, 'a', 'b')])

  def test_redirected(self):
    """Test redirected."""
    self.mock.get_access.return_value = access.UserAccess.Redirected
    self.assertFalse(access.has_access(True, 'a', 'b'))
    self.mock.get_access.assert_has_calls([mock.call(True, 'a', 'b')])


@test_utils.with_cloud_emulators('datastore')
class CanUserAccessTestcaseTest(unittest.TestCase):
  """Test can_user_access_testcase."""

  def setUp(self):
    test_helpers.patch(self, [
        'libs.access._is_domain_allowed',
        'libs.auth.get_current_user',
        'clusterfuzz._internal.config.db_config.get',
        'libs.issue_management.issue_tracker.IssueTracker.get_original_issue',
        'libs.issue_management.issue_tracker_utils.'
        'get_issue_tracker_for_testcase',
        'libs.issue_management.monorail.issue_tracker_manager.'
        'IssueTrackerManager',
    ])
    self.itm = issue_tracker_manager.IssueTrackerManager('test')
    self.itm.project_name = 'test-project'
    self.mock.get_issue_tracker_for_testcase.return_value = (
        monorail.IssueTracker(self.itm))
    self.get_issue = self.itm.get_issue

    self.email = 'test@test.com'
    self.mock.get_current_user.return_value = auth.User(self.email)

    self.bug = issue.Issue()
    self.bug.id = 1234
    self.bug.itm = self.itm
    self.original_bug = issue.Issue()
    self.original_bug.id = 5678
    self.original_bug.itm = self.itm

    self.testcase = data_types.Testcase()

    self.mock.get.return_value = (
        data_types.Config(relax_testcase_restrictions=True))
    self.mock._is_domain_allowed.return_value = False

  def test_allowed(self):
    """Ensure it is true when check_user_access allows for a specific
       job_type."""
    data_types.ExternalUserPermission(
        email=self.email,
        entity_name='job',
        entity_kind=data_types.PermissionEntityKind.JOB,
        auto_cc=data_types.AutoCCType.ALL).put()

    self.testcase.job_type = 'job'
    self.testcase.fuzzer_name = 'fuzzer'
    self.testcase.security_flag = True
    self.assertTrue(access.can_user_access_testcase(self.testcase))

  def _test_bug_access(self):
    self.get_issue.return_value = self.bug

    self.testcase.bug_information = '1234'
    self.assertTrue(access.can_user_access_testcase(self.testcase))

    self.get_issue.assert_has_calls([mock.call(1234)])

  def test_no_bug(self):
    """Ensure it is false when there's no bug."""
    self.testcase.bug_information = ''
    self.assertFalse(access.can_user_access_testcase(self.testcase))

    self.get_issue.assert_has_calls([])

  def test_invalid_bug(self):
    """Ensure it is false when testcase's bug is invalid."""
    self.get_issue.return_value = None

    self.testcase.bug_information = '1234'
    self.assertFalse(access.can_user_access_testcase(self.testcase))

    self.get_issue.assert_has_calls([mock.call(1234)])

  def test_allowed_because_of_cc(self):
    """Ensure it is allowed because the user is CC."""
    self.bug.add_cc(self.email.capitalize())
    self._test_bug_access()

  def test_allowed_because_of_owner(self):
    """Ensure it is allowed because the user is the owner."""
    self.bug.owner = self.email.capitalize()
    self._test_bug_access()

    self.mock.get.return_value = (
        data_types.Config(relax_testcase_restrictions=False))
    self._test_bug_access()

  def test_allowed_because_of_owner_group(self):
    """Ensure it is allowed because the user is the owner of the group bug."""
    self.bug.owner = self.email.capitalize()
    self.get_issue.return_value = self.bug

    self.testcase.bug_information = None
    self.testcase.group_bug_information = 1234

    self.assertTrue(access.can_user_access_testcase(self.testcase))
    self.get_issue.assert_has_calls([mock.call(1234)])

  def test_allowed_because_of_reporter(self):
    """Ensure it is allowed because the user is the reporter."""
    self.bug.reporter = self.email.capitalize()
    self._test_bug_access()

  def test_allowed_because_of_domain_allowed(self):
    """Ensure it is true when user has bug access and user's email is on the
      domain list but the relaxation is not enabled."""
    self.mock._is_domain_allowed.return_value = True
    self.testcase.security_flag = True
    self.mock.get.return_value = (
        data_types.Config(relax_testcase_restrictions=False))
    self.bug.add_cc(self.email)
    self._test_bug_access()

  def test_allowed_because_of_uploader(self):
    """Ensure it is allowed because the user is the uploader."""
    self.testcase.uploader_email = 'test@test.com'
    self.testcase.security_flag = True

    self.assertTrue(access.can_user_access_testcase(self.testcase))

  def test_allowed_because_of_owner_in_original_issue(self):
    """Ensure it is allowed because the user is the owner of original issue."""
    self.bug.merged_into = 5678
    self.bug.merged_into_project = 'test-project'
    self.original_bug.owner = self.email
    self.mock.get_original_issue.return_value = monorail.Issue(
        self.original_bug)
    self._test_bug_access()

    self.mock.get.return_value = (
        data_types.Config(relax_testcase_restrictions=False))
    self._test_bug_access()

  def test_allowed_security_bug_access_for_domain_user(self):
    """Ensure that a domain user can access a security bug if restrictions
    are relaxed in configuration."""
    self.mock._is_domain_allowed.return_value = True
    self.testcase.security_flag = True
    self.mock.get.return_value = (
        data_types.Config(relax_security_bug_restrictions=True))

    self.assertTrue(access.can_user_access_testcase(self.testcase))

  def test_denied_security_bug_access_for_domain_user(self):
    """Ensure that a domain user can't access a security bug in default
    configuration."""
    self.mock._is_domain_allowed.return_value = True
    self.testcase.security_flag = True
    self.mock.get.return_value = (
        data_types.Config(relax_security_bug_restrictions=False))

    self.assertFalse(access.can_user_access_testcase(self.testcase))

  def test_denied_no_access(self):
    """Ensure it is false when user has bug access but the relaxation is not
      enabled and user's email is not on the allowed domain list."""
    self.mock._is_domain_allowed.return_value = False
    self.mock.get.return_value = (
        data_types.Config(relax_testcase_restrictions=False))
    self.testcase.bug_information = '1234'
    self.get_issue.return_value = self.bug

    self.bug.add_cc(self.email)
    self.bug.reporter = self.email
    self.bug.owner = ''

    self.assertFalse(access.can_user_access_testcase(self.testcase))

  def test_deny_no_access_and_no_bug_access(self):
    """Ensure it is false when user has no access and no bug access."""
    self.mock._is_domain_allowed.return_value = False
    test_issue = issue.Issue()
    test_issue.itm = self.itm
    self.get_issue.return_value = test_issue

    self.testcase.bug_information = '1234'
    self.assertFalse(access.can_user_access_testcase(self.testcase))

    self.get_issue.assert_has_calls([mock.call(1234)])


@test_utils.with_cloud_emulators('datastore')
class CheckAccessAndGetTestcase(unittest.TestCase):
  """Test check_access_and_get_testcase."""

  def setUp(self):
    test_helpers.patch(self, [
        'libs.access.has_access',
        'libs.helpers.get_user_email',
    ])
    self.mock.get_user_email.return_value = 'user@example.com'

    token = data_types.CSRFToken()
    token.user_email = self.mock.get_user_email.return_value
    token.put()

    self.testcase = data_types.Testcase()
    self.testcase.put()

  def test_not_logged_in(self):
    """Test not logged in."""
    self.mock.get_user_email.return_value = ''
    with self.assertRaises(helpers.UnauthorizedException):
      access.check_access_and_get_testcase(self.testcase.key.id())

  def test_privileged(self):
    """Test being privileged (never get locked)."""
    self.mock.has_access.return_value = True
    access.check_access_and_get_testcase(self.testcase.key.id())

    with self.assertRaises(helpers.EarlyExitException) as cm:
      access.check_access_and_get_testcase(self.testcase.key.id() + 1)
    self.assertEqual(404, cm.exception.status)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fast minimizer that attempts to remove tokens grouped in chunks."""

import functools

from . import errors
from . import minimizer
from . import utils


class ChunkMinimizer(minimizer.Minimizer):
  """Minimizer to replicate the old ClusterFuzz minimization strategy."""

  def __init__(self, *args, **kwargs):
    self.chunk_sizes = self._handle_constructor_argument(
        'chunk_sizes', kwargs, default=[10, 4, 1])
    minimizer.Minimizer.__init__(self, *args, **kwargs)

  def _execute(self, data):
    """Minimize |data| using the algorithm from CF (but backwards)."""
    testcase = minimizer.Testcase(data, self)
    if not self.validate_tokenizer(data, testcase):
      raise errors.TokenizationFailureError('Chunk Minimizer')

    for lines_to_remove in self.chunk_sizes:
      remaining_tokens = testcase.get_required_token_indices()
      for end_index in range(len(remaining_tokens), 0, -lines_to_remove):
        start_index = max(0, end_index - lines_to_remove)
        hypothesis = remaining_tokens[start_index:end_index]
        testcase.prepare_test(hypothesis)

      testcase.process()

    return testcase

  @staticmethod
  def run(data, thread_count=minimizer.DEFAULT_THREAD_COUNT, file_extension=''):
    """Minimize |data| using the old strategy from CF."""
    minimizer_round_1 = ChunkMinimizer(
        utils.test,
        max_threads=thread_count,
        tokenizer=utils.tokenize,
        token_combiner=utils.token_combiner,
        chunk_sizes=[80, 40, 20],
        file_extension=file_extension)

    full_tokenizer = functools.partial(utils.tokenize, level=1)
    minimizer_round_2 = ChunkMinimizer(
        utils.test,
        max_threads=thread_count,
        tokenizer=full_tokenizer,
        token_combiner=utils.token_combiner,
        chunk_sizes=[10, 4, 1],
        file_extension=file_extension)

    result = minimizer_round_1.minimize(data)
    result = minimizer_round_2.minimize(result)
    return result



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""update_from_trunk tests."""
import unittest

import flask
import webtest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers.testcase_detail import update_from_trunk
from libs import form


@test_utils.with_cloud_emulators('datastore')
class HandlerTest(unittest.TestCase):
  """Test Handler."""

  def setUp(self):
    test_helpers.patch(self, [
        'clusterfuzz._internal.base.tasks.add_task',
        'clusterfuzz._internal.base.tasks.queue_for_job',
        'libs.auth.get_current_user',
        'handlers.testcase_detail.show.get_testcase_detail',
        'libs.access.check_access_and_get_testcase',
    ])
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=update_from_trunk.Handler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

    self.testcase = data_types.Testcase(queue='old-queue')
    self.testcase.put()
    self.mock.check_access_and_get_testcase.return_value = self.testcase
    self.mock.get_testcase_detail.return_value = {'testcase': 'yes'}
    self.mock.get_current_user().email = 'test@user.com'

  def test_succeed(self):
    """Update from trunk"""
    self.mock.queue_for_job.return_value = 'jobs-suffix'

    self.testcase.crash_stacktrace = 'Random'
    self.testcase.job_type = 'job'
    self.testcase.put()

    resp = self.app.post_json(
        '/', {
            'testcaseId': self.testcase.key.id(),
            'csrf_token': form.generate_csrf_token(),
        })

    self.assertEqual(200, resp.status_int)
    self.assertEqual('yes', resp.json['testcase'])
    self.mock.add_task.assert_called_once_with(
        'variant', self.testcase.key.id(), 'job', queue='jobs-suffix')

    testcase = self.testcase.key.get()
    self.assertEqual('Pending', testcase.last_tested_crash_stacktrace)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Trusted host/untrusted instance configuration."""

PORT = 9001

INITIAL_CONNECT_TIMEOUT_SECONDS = 5 * 60
RECONNECT_TIMEOUT_SECONDS = 30

FILE_TRANSFER_CHUNK_SIZE = 4096

HEARTBEAT_INTERVAL_SECONDS = 60
HEARTBEAT_TIMEOUT_SECONDS = 15

GET_STATUS_TIMEOUT_SECONDS = 15
UPDATE_SOURCE_TIMEOUT_SECONDS = 15

RPC_RETRY_ATTEMPTS = 1

GRPC_OPTIONS = (
    ('grpc.max_send_message_length', -1),
    ('grpc.max_receive_message_length', -1),
    ('grpc.max_metadata_size', 32 * 1024 * 1024),
)

NUM_WORKER_THREADS = 4
